# !/usr/bin/python3
"""
CCP4 Parser (pdb_eda.ccp4)
-------------------------------------------------------

This module provides methods to read and parse the CCP4 format files, returning ccp4 objects.
Format details of ccp4 can be found in http://www.ccp4.ac.uk/html/maplib.html.
"""

import warnings
import struct

import urllib.request
import numpy as np

try:
    from . import cutils as utils
except ImportError:
    from . import utils

urlPrefix = "http://www.ebi.ac.uk/pdbe/coordinates/files/"
urlSuffix = ".ccp4"


def readFromPDBID(pdbid, verbose=False):
    """Creates :class:`pdb_eda.ccp4.DensityMatrix` object.

    :param pdbid: PDB entry ID.
    :type pdbid: :py:class:`str`
    :param verbose: verbose mode, defaults to :py:obj:`False`
    :type verbose: :py:class:`bool`

    :return: densityMatrix
    :rtype: :class:`pdb_eda.ccp4.DensityMatrix`
    """
    return readFromURL(urlPrefix + pdbid.lower() + urlSuffix, pdbid, verbose)


def readFromURL(url, pdbid=None, verbose=False):
    """Creates :class:`pdb_eda.ccp4.DensityMatrix` object.

    :param url:
    :type url: :py:class:`str`
    :param pdbid: PDB entry ID.
    :type pdbid: :py:class:`str`, optional
    :param verbose: verbose mode, defaults to :py:obj:`False`
    :type verbose: :py:class:`bool`

    :return: densityMatrix
    :rtype: :class:`pdb_eda.ccp4.DensityMatrix`
    """
    if not pdbid:
        pdbid = url
    with urllib.request.urlopen(url) as urlHandle:
        return parse(urlHandle, pdbid, verbose)


def read(ccp4Filename, pdbid=None, verbose=False):
    """Creates :class:`pdb_eda.ccp4.DensityMatrix` object.

    :param ccp4Filename: .ccp4 filename including path.
    :type ccp4Filename: :py:class:`str`
    :param pdbid: PDB entry ID.
    :type pdbid: :py:class:`str`, optional
    :param verbose: verbose mode, defaults to :py:obj:`False`
    :type verbose: :py:class:`bool`

    :return: densityMatrix
    :rtype: :class:`pdb_eda.ccp4.DensityMatrix`
    """
    if not pdbid:
        pdbid = ccp4Filename
    with open(ccp4Filename, "rb") as fileHandle:
        return parse(fileHandle, pdbid, verbose)


def parse(handle, pdbid, verbose=False):
    """Creates :class:`pdb_eda.ccp4.DensityMatrix` object.

    :param handle: an I/O handle for .ccp4 file.
    :type handle: :class:`io.IOBase`
    :param pdbid: PDB entry ID.
    :type pdbid: :py:class:`str`
    :param verbose: verbose mode, defaults to :py:obj:`False`
    :type verbose: :py:class:`bool`

    :return: densityMatrix
    :rtype: :class:`pdb_eda.ccp4.DensityMatrix`
    """
    header = DensityHeader.fromFileHeader(handle.read(1024))
    endian = header.endian
    dataBuffer = handle.read()

    # Sanity check on file sizes
    if len(dataBuffer) != header.symmetryBytes + header.mapSize:
        assert header.symmetryBytes == 0 | len(
            dataBuffer) != header.mapSize, "Error: File contains suspicious symmetry records"
        assert header.mapSize == 0 | len(dataBuffer) != header.symmetryBytes, "Error: File contains no map data"
        assert len(dataBuffer) > header.symmetryBytes + header.mapSize, "Error: contains incomplete data"
        assert len(dataBuffer) < header.symmetryBytes + header.mapSize, "Error: File contains larger than expected data"

    assert header.xlength != 0.0 or header.ylength != 0.0 or header.zlength != 0.0, "Error: Cell dimensions are all 0, Map file will not align with other structures"

    if header.nintervalX == 0 & header.ncrs[0] > 0:
        header.nintervalX = header.ncrs[0] - 1
        if verbose: warnings.warn("Fixed number of X interval")
    if header.nintervalY == 0 & header.ncrs[1] > 0:
        header.nintervalY = header.ncrs[1] - 1
        if verbose: warnings.warn("Fixed number of Y interval")
    if header.nintervalZ == 0 & header.ncrs[2] > 0:
        header.nintervalZ = header.ncrs[2] - 1
        if verbose: warnings.warn("Fixed number of Z interval.")

    if header.col2xyz == 0 & header.row2xyz == 0 & header.sec2xyz == 0:
        header.col2xyz = 1
        header.row2xyz = 2
        header.sec2xyz = 3
        if verbose: warnings.warn("Mappings from column/row/section to xyz are all 0, set to 1, 2, 3 instead.")

    header.symmetry = dataBuffer[0:header.symmetryBytes]
    mapData = dataBuffer[header.symmetryBytes:len(dataBuffer)]

    numBytes = int(len(mapData) / 4)
    densities = struct.unpack(endian + numBytes * 'f', mapData)
    origin = header.origin

    return DensityMatrix(header, origin, densities, pdbid)


class DensityHeader(object):
    """:class:`pdb_eda.ccp4.DensityHeader` that stores information about ccp4 header."""

    @classmethod
    def fromFileHeader(cls, fileHeader):
        """RETURNS :class:`pdb_eda.ccp4.DensityHeader` object given the fileHeader.

        :param fileHeader: ccp4 file header.
        :type fileHeader: :py:class:`bytes`

        :return: densityHeader
        :rtype: :class:`pdb_eda.ccp4.DensityHeader`
        """

        # Test for endianness
        mode = int.from_bytes(fileHeader[12:16], byteorder='little')
        endian = '<' if 0 <= mode <= 6 else '>'

        # Header
        headerFormat = endian + 10 * 'i' + 6 * 'f' + 3 * 'i' + 3 * 'f' + 3 * 'i' + 27 * 'f' + 4 * 'c' + 'ifi'
        headerTuple = struct.unpack(headerFormat, fileHeader[:224])
        #print(headerTuple)
        labels = fileHeader[224:]  # Labels in header
        labels = labels.replace(b' ', b'')

        header = DensityHeader(headerTuple, labels, endian)
        return header

    def __init__(self, headerTuple, labels, endian):
        """Initialize the DensityHeader object, assign values to data members accordingly, and calculate some metrics that will be used frequently.

        :param headerTuple: The ccp4 header information (excluding labels) in a tuple.
        :type headerTuple: :py:class:`tuple`
        :param labels: The labels field in a ccp4 header.
        :type labels: :py:class:`bytes`
        :param endian: The endianness of the file.
        :type endian: :py:class:`str`
        """
        self.ncrs = headerTuple[0:3]
        #Number of Columns    (fastest changing in map)
        #Number of Rows
        #Number of Sections   (slowest changing in map)

        self.mode = headerTuple[3]
        self.endian = endian
        #Data type
        #    0 = envelope stored as signed bytes (from -128 lowest to 127 highest)
        #    1 = Image     stored as Integer*2
        #    2 = Image     stored as Reals
        #    3 = Transform stored as Complex Integer*2
        #    4 = Transform stored as Complex Reals
        #    5 == 0

        #    Note: Mode 2 is the normal mode used in the CCP4 programs. Other modes than 2 and 0
        #        may NOT WORK

        self.crsStart = headerTuple[4:7]  # Number of first COLUMN, ROW, and SECTION in map
        self.nintervalX = headerTuple[7]  # Number of intervals along X
        self.nintervalY = headerTuple[8]  # Number of intervals along Y
        self.nintervalZ = headerTuple[9]  # Number of intervals along Z
        self.xlength = headerTuple[10]  # Cell Dimensions (Angstroms)
        self.ylength = headerTuple[11]  # ''
        self.zlength = headerTuple[12]  # ''
        self.alpha = headerTuple[13]  # Cell Angles     (Degrees)
        self.beta = headerTuple[14]  # ''
        self.gamma = headerTuple[15]  # ''
        self.col2xyz = headerTuple[16]  # Which axis corresponds to Cols.  (1,2,3 for X,Y,Z)
        self.row2xyz = headerTuple[17]  # Which axis corresponds to Rows   (1,2,3 for X,Y,Z)
        self.sec2xyz = headerTuple[18]  # Which axis corresponds to Sects. (1,2,3 for X,Y,Z)
        self.densityMin = headerTuple[19]  # Minimum density value
        self.densityMax = headerTuple[20]  # Maximum density value
        self.densityMean = headerTuple[21]  # Mean    density value    (Average)
        self.spaceGroup = headerTuple[22]  # Space group number
        self.symmetryBytes = headerTuple[23]  # Number of bytes used for storing symmetry operators
        self.skewFlag = headerTuple[24]  # Flag for skew transformation, =0 none, =1 if foll
        self.skewMat = headerTuple[25:34]  # Skew matrix S (in order S11, S12, S13, S21 etc) if LSKFLG .ne. 0.
        self.skewTrans = headerTuple[34:37]
        #Skew translation t if LSKFLG .ne. 0.
        #            Skew transformation is from standard orthogonal
        #            coordinate frame (as used for atoms) to orthogonal
        #            map frame, as: Xo(map) = S * (Xo(atoms) - t)

        self.futureUse = headerTuple[37:49]
        #(some of these are used by the MSUBSX routines in MAPBRICK, MAPCONT and FRODO) (all set to zero by default)
        self.originEM = headerTuple[49:52]
        #Use ORIGIN records rather than old crsStart records as in http://www2.mrc-lmb.cam.ac.uk/image2000.html
        #The ORIGIN field is only used by the EM community, and has undefined meaning for non-orthogonal maps and/or
        #non-cubic voxels, etc.

        self.mapChar = headerTuple[52:56]  # Character string 'MAP ' to identify file type
        self.machineStamp = headerTuple[56]  # Machine stamp indicating the machine type which wrote file
        self.rmsd = headerTuple[57]  # Rms deviation of map from mean density
        self.nLabel = headerTuple[58]  # Number of labels being used
        self.labels = labels

        self.mapSize = self.ncrs[0] * self.ncrs[1] * self.ncrs[2] * 4
        self.xyzLength = [self.xlength, self.ylength, self.zlength]
        self.xyzInterval = [self.nintervalX, self.nintervalY, self.nintervalZ]
        self.gridLength = [x/y for x, y in zip(self.xyzLength, self.xyzInterval)]

        indices = [0, 0, 0]
        indices[self.col2xyz - 1] = 0
        indices[self.row2xyz - 1] = 1
        indices[self.sec2xyz - 1] = 2
        self.map2xyz = indices
        self.map2crs = [self.col2xyz - 1, self.row2xyz - 1, self.sec2xyz - 1]

        self.crsInterval = [self.xyzInterval[self.map2crs[ind]] for ind in range(3)]


        alpha = np.pi / 180 * self.alpha
        beta = np.pi / 180 * self.beta
        gamma = np.pi / 180 * self.gamma
        self.unitVolume = self.xlength * self.ylength * self.zlength / self.nintervalX / self.nintervalY / self.nintervalZ * \
                          np.sqrt(1 - np.cos(alpha) ** 2 - np.cos(beta) ** 2 - np.cos(gamma) ** 2 + 2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma))

        ## A reusable part in the cell volumn calculation
        temp = np.sqrt(1 - np.cos(alpha) ** 2 - np.cos(beta) ** 2 - np.cos(gamma) ** 2 + 2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma))
        self.orthoMat = [[self.xlength, self.ylength * np.cos(gamma), self.zlength * np.cos(beta)],
                         [0, self.ylength * np.sin(gamma), self.zlength * (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma)],
                         [0, 0, self.zlength * temp / np.sin(gamma)]]

        self.deOrthoMat = np.linalg.inv(self.orthoMat)
        self.deOrthoMat[abs(self.deOrthoMat) < 1e-10] = 0.0

        #self.deOrthoMat = [[1/self.xlength, - np.cos(gamma) / np.sin(gamma) / self.xlength,
        #                    (np.cos(gamma) * (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma) - np.cos(beta) * np.sin(gamma)) / self.xlength / temp],
        #                   [0, 1/np.sin(gamma)/self.ylength, - (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma) / self.ylength / temp],
        #                   [0, 0, np.sin(gamma) / self.zlength / temp]]

        self.origin = self._calculateOrigin()

        ncrs = [i for i in self.ncrs]
        if self.xyzInterval[self.col2xyz - 1] < self.ncrs[0]:
            ncrs[0] = self.xyzInterval[self.col2xyz - 1]
        if self.xyzInterval[self.row2xyz - 1] < self.ncrs[1]:
            ncrs[1] = self.xyzInterval[self.row2xyz - 1]
        if self.xyzInterval[self.sec2xyz - 1] < self.ncrs[2]:
            ncrs[2] = self.xyzInterval[self.sec2xyz - 1]
        self.uniqueNcrs = ncrs


    def _calculateOrigin(self):
        """Calculate the xyz coordinates from the header information.

        :return: xyz coordinates.
        :rtype: :py:class:`list` of :py:class:`float`.
        """
        # Orthogonalization matrix for calculation between fractional coordinates and orthogonal coordinates
        # Formula based on 'Biomolecular Crystallography' by Bernhard Rupp, p233

        if self.futureUse[-3] == 0.0 and self.futureUse[-2] == 0.0 and self.futureUse[-1] == 0.0:
            origin = np.dot(self.orthoMat, [self.crsStart[self.map2xyz[i]] / self.xyzInterval[i] for i in range(3)])
        else:
            origin = [self.originEM[i] for i in range(3)]

        return origin

    def xyz2crsCoord(self, xyzCoord):
        """Convert the xyz coordinates into crs coordinates.

        :param xyzCoord: xyz coordinates.
        :type xyzCoord: :py:class:`list` of :py:class:`float`

        :return: crs coordinates.
        :rtype: A :py:class:`list` of :py:class:`int`.
        """
        if self.alpha == self.beta == self.gamma == 90:
            crsGridPos = [int(round((xyzCoord[i] - self.origin[i]) / self.gridLength[i])) for i in range(3)]
        else:
            fraction = np.dot(self.deOrthoMat, xyzCoord)
            crsGridPos = [int(round(fraction[i] * self.xyzInterval[i])) - self.crsStart[self.map2xyz[i]] for i in range(3)]
        return [crsGridPos[self.map2crs[i]] for i in range(3)]

    def crs2xyzCoord(self, crsCoord):
        """Convert the crs coordinates into xyz coordinates.

        :param crsCoord: crs coordinates.
        :type crsCoord: A :py:obj:`list` of :py:class:`int`

        :return: xyz coordinates.
        :rtype: A :py:class:`list` of :py:class:`float`.
        """
        if self.alpha == self.beta == self.gamma == 90:
            return [crsCoord[self.map2xyz[i]] * self.gridLength[i] + self.origin[i] for i in range(3)]
        else:
            return np.dot(self.orthoMat, [(crsCoord[self.map2xyz[i]] + self.crsStart[self.map2xyz[i]]) / self.xyzInterval[i] for i in range(3)])


class DensityMatrix:
    """:class:`pdb_eda.ccp4.DensityMatrix` that stores data and methods of a ccp4 file."""

    def __init__(self, header, origin, density, pdbid):
        """Initialize the :class:`pdb_eda.ccp4.DensityMatrix` object.

        :param header:
        :type header: :class:`pdb_eda.ccp4.DensityHeader`
        :param origin: the xyz coordinates of the origin of the first number of the density data.
        :type origin: :py:class:`list`
        :param density: the density data as a 1-d list.
        :type density: :py:class:`tuple`
        :param pdbid: PDB entry ID
        :type pdbid: :py:class:`str`
        """
        self.pdbid = pdbid
        self.header = header
        self.origin = origin
        self.densityArray = density
        self.density = np.array(density).reshape(header.ncrs[2], header.ncrs[1], header.ncrs[0])
        self._meanDensity = None
        self._stdDensity = None
        self._totalAbsDensity = {}

    @property
    def meanDensity(self):
        """Returns mean of the density.

        :return: mean
        :rtype: :py:class:`float`
        """
        if self._meanDensity == None:
            self._meanDensity = np.mean(self.densityArray)
        return self._meanDensity

    @property
    def stdDensity(self):
        """Returns the standard deviation of the density.

        :return: std
        :rtype: :py:class:`float`
        """
        if self._stdDensity == None:
            self._stdDensity = np.std(self.densityArray)
        return self._stdDensity

    def getTotalAbsDensity(self, densityCutoff):
        """Returns total absolute Density above a densityCutoff

        :param densityCutoff:
        :type densityCutoff: :py:class:`float`

        :return: totalAbsDensity
        :rtype: :py:class:`float`
        """
        if densityCutoff not in self._totalAbsDensity:
            self._totalAbsDensity[densityCutoff] = utils.sumOfAbs(self.densityArray, densityCutoff)
        return self._totalAbsDensity[densityCutoff]

    def getPointDensityFromCrs(self, crsCoord):
        """Get the density of a point.

        :param crsCoord: crs coordinates.
        :type crsCoord: A :py:class:`list` of :py:class:`int`

        :return: pointDensity
        :rtype: :py:class:`float`
        """
        return utils.getPointDensityFromCrs(self,crsCoord)

    def getPointDensityFromXyz(self, xyzCoord):
        """Get the density of a point.

        :param xyzCoord: xyz coordinates.
        :type xyzCoord: A :py:class:`list` of :py:class:`float`

        :return: pointDensity
        :rtype: :py:class:`float`
        """
        return utils.getPointDensityFromCrs(self, self.header.xyz2crsCoord(xyzCoord))

    def getSphereCrsFromXyz(self, xyzCoord, radius, densityCutoff=0):
        """Calculate a list of crs coordinates that within a given distance of a point.

        :param xyzCoord: xyz coordinates.
        :type xyzCoord: A :py:class:`list` of :py:class:`float`
        :param radius:
        :type radius: :py:class:`float`
        :param densityCutoff: a density cutoff for all the points wants to be included, defaults to 0
                Default 0 means include every point within the radius.
                If cutoff < 0, include only points with density < cutoff.
                If cutoff > 0, include only points with density > cutoff.
        :type densityCutoff: :py:class:`float`

        :return: crsList
        :rtype: :py:class:`list`
        """
        return utils.getSphereCrsFromXyz(self,xyzCoord,radius,densityCutoff)

    def getTotalDensityFromXyz(self, xyzCoord, radius, densityCutoff=0):
        """Calculate the total density of a sphere.

        :param xyzCoord: xyz coordinates.
        :type xyzCoord: :py:class:`list` of :py:class:`float`
        :param radius:
        :type radius: :py:class:`float`
        :param densityCutoff: a density cutoff for all the points to include, defaults to 0
                Default 0 means include every point within the radius.
                If cutoff < 0, include only points with density < cutoff.
                If cutoff > 0, include only points with density > cutoff.
        :type densityCutoff: :py:class:`float`

        :return: density
        :rtype: :py:class:`float`
        """
        crsCoordList = utils.getSphereCrsFromXyz(self, xyzCoord, radius, densityCutoff)
        return sum(utils.getPointDensityFromCrs(self, crs) for crs in crsCoordList)

    def findAberrantBlobs(self, xyzCoords, radius, densityCutoff=0):
        """Within a given radius, find and aggregate all neighbouring aberrant points into blobs (red/green meshes).

        :param xyzCoords: single xyz coordinate or a list of xyz coordinates.
        :type xyzCoords: :py:class:`list`
        :param radius:
        :type radius: :py:class:`float`
        :param densityCutoff: A density cutoff for all the points wants to be included, defaults to 0
                Default 0 means include every point within the radius.
                If cutoff < 0, include only points with density < cutoff.
                If cutoff > 0, include only points with density > cutoff.
        :type densityCutoff: :py:class:`float`

        :return: blobList  of aberrant blobs described by their xyz centroid, total density, and volume.
        :rtype: :py:class:`list` of :class:`pdb_eda.ccp4.DensityBlob` objects.
        """
        if not isinstance(xyzCoords[0], (np.floating, float)): # test if xyzCoords is a single xyzCoord or a list of them.
            if len(xyzCoords) > 1:
                crsCoordList = list(utils.getSphereCrsFromXyzList(self, xyzCoords, radius, densityCutoff))
            else:
                crsCoordList = utils.getSphereCrsFromXyz(self, xyzCoords[0], radius, densityCutoff)
        else:
            crsCoordList = utils.getSphereCrsFromXyz(self, xyzCoords, radius, densityCutoff)

        return self.createBlobList(crsCoordList)

    def createFullBlobList(self, cutoff):
        """Aggregate the density map into positive (green or blue) or negative (red) blobs.

        :param cutoff: density cutoff to use to filter voxels.
        :type cutoff: :py:class:`float`

        :return blobList: list of DensityBlobs
        :rtype: :py:class:`list` of :class:`pdb_eda.ccp4.DensityBlob` objects.
        """
        crsList = utils.createFullCrsList(self, cutoff)
        return self.createBlobList(crsList) if crsList != None else None

    def createBlobList(self, crsList):
        """Calculates a list of blobs from a given crsList.

        :param crsList: a crs list.
        :type crsList: :py:class:`list`, :py:class:`set`

        :return: blobList
        :rtype: :py:class:`list` of :class:`pdb_eda.ccp4.DensityBlob` objects.
        """
        crsLists = utils.createCrsLists(crsList)
        return [ DensityBlob.fromCrsList(crs_list, self) for crs_list in crsLists ]


class DensityBlob:
    """:class:`pdb_eda.ccp4.DensityBlob` that stores data and methods of a electron density blob."""

    def __init__(self, centroid, coordCenter, totalDensity, volume, crsList, densityMatrix, atoms=None):
        """Initialize a :class:`pdb_eda.ccp4.DensityBlob` object.

        :param centroid: the centroid of the blob.
        :type centroid: :py:class:`list`
        :param totalDensity: the totalDensity of the blob.
        :type totalDensity: :py:class:`float`
        :param volume: the volume of the blob = number of density units * unit volumes.
        :type volume: :py:class:`float`
        :param crsList: the crs list of the blob.
        :type crsList: :py:class:`list`
        :param densityMatrix: the entire density map that the blob belongs to.
        :type densityMatrix: `pdb_eda.ccp4.DensityMatrix`
        :param atoms: list of atoms for the blob.
        :type atoms: :py:class:`list`, optional

        :return: densityBlob
        :rtype: :class:`pdb_eda.ccp4.DensityBlob`
        """
        self.centroid = centroid
        self.coordCenter = coordCenter
        self.totalDensity = totalDensity
        self.volume = volume
        self.crsList = {tuple(crs) for crs in crsList}
        self.densityMatrix = densityMatrix
        self.atoms = [] if not atoms else atoms

    @property
    def validCrs(self):
        return utils.testValidCrsList(self.densityMatrix, self.crsList)

    @staticmethod
    def fromCrsList(crsList, densityMatrix):
        """The creator of a A :class:`pdb_eda.ccp4.DensityBlob` object.

        :param crsList: the crs list of the blob.
        :type crsList: :py:class:`list`
        :param densityMatrix: the 3-d density matrix to use for calculating centroid etc, so the object does not have to have a density list data member.
        :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`

        :return: densityBlob
        :rtype: :class:`pdb_eda.ccp4.DensityBlob`
        """
        weights = [0, 0, 0]
        totalDensity = 0
        for i, point in enumerate(crsList):
            density = utils.getPointDensityFromCrs(densityMatrix, point)
            pointXYZ = densityMatrix.header.crs2xyzCoord(point)
            weights = [weights[i] + density * pointXYZ[i] for i in range(3)]
            totalDensity += density

        centroidXYZ = [weight / totalDensity for weight in weights]
        npoints = len(crsList)
        coordCenter = [sum(k) / npoints for k in zip(*[densityMatrix.header.crs2xyzCoord(crs) for crs in crsList])]
        return DensityBlob(centroidXYZ, coordCenter, totalDensity, densityMatrix.header.unitVolume * len(crsList), crsList, densityMatrix)


    def __eq__(self, otherBlob):
        """Check if two blobs are the same, and overwrite the '==' operator for the :class:`pdb_eda.ccp4.DensityBlob` object.

        :param otherBlob:
        :type otherBlob: :class:`pdb_eda.ccp4.DensityBlob`

        :return: bool
        :rtype: :py:class`bool`
        """
        if abs(self.volume - otherBlob.volume) >= 1e-6: return False
        if abs(self.totalDensity - otherBlob.totalDensity) >= 1e-6: return False
        for i in range(0, 3):
            if abs(self.centroid[i] - otherBlob.centroid[i]) >= 1e-6: return False

        return True

    def testOverlap(self, otherBlob):
        """Check if two blobs overlaps or right next to each other.

        :param otherBlob:
        :type otherBlob: :class:`pdb_eda.ccp4.DensityBlob`

        :return: bool
        :rtype: :py:class`bool`
        """
        return utils.testOverlap(self, otherBlob)

    def merge(self, otherBlob):
        """Merge the given blob into the original blob.

        :param otherBlob:
        :type otherBlob: :class:`pdb_eda.ccp4.DensityBlob`
        """
        self.crsList.update(otherBlob.crsList)
        atoms = self.atoms + [atom for atom in otherBlob.atoms if atom not in self.atoms]
        newBlob = DensityBlob.fromCrsList(self.crsList, self.densityMatrix)

        self.__dict__.update(newBlob.__dict__)
        self.atoms = atoms

    def clone(self):
        """Returns a copy of the density blob.

        :return: densityBlob
        :rtype: :class:`pdb_eda.ccp4.DensityBlob`
        """
        return DensityBlob(self.centroid,self.coordCenter,self.totalDensity,self.volume,self.crsList,self.densityMatrix,self.atoms.copy())




"""
Utilities (pdb_eda.utils)
-------------------------

Contains low-level functions used in pdb_eda.ccp4 and pdb_eda.densityAnalysis.
Should not be used, but is a fallback if cutils.pyx cannot be cythonized.
The cythonized version provide a 3- to 4-fold improvement in execution performance.
"""
def testOverlap(selfBlob, otherBlob):
    """Check if two blobs overlaps or right next to each other.

    :param selfBlob:
    :type selfBlob: :class:`pdb_eda.ccp4.DensityBlob`
    :param otherBlob:
    :type otherBlob: :class:`pdb_eda.ccp4.DensityBlob`

    :return: bool
    :rtype: :py:class:`bool`
    """
    if len(selfBlob.crsList) > len(otherBlob.crsList):
        return True if any(-1 <= x[0] - y[0] <= 1 and -1 <= x[1] - y[1] <= 1 and -1 <= x[2] - y[2] <= 1 for x in selfBlob.crsList for y in otherBlob.crsList) else False
    else:
        return True if any(-1 <= x[0] - y[0] <= 1 and -1 <= x[1] - y[1] <= 1 and -1 <= x[2] - y[2] <= 1 for x in otherBlob.crsList for y in selfBlob.crsList) else False


def sumOfAbs(array, cutoff):
    """Return sum of absolute values above a cutoff.

    :param array:
    :type array: :class:`collections.abc.Iterable`
    :param cutoff:
    :type cutoff: :py:class:`float`

    :return: value
    :rtype: :py:class:`float`
    """
    return sum(abs(value) for value in array if abs(value) > cutoff)

import numpy as np
import scipy.spatial
dcutoff = np.sqrt(3)  ## the points are considered to be adjacent if one is in the one layer outer box with the other one in the center
def createCrsLists(crsList):
    """Calculates a list of crsLists from a given crsList.
    This is a preparation step for creating blobs.

    :param crsList: a crs list.
    :type crsList: :py:class:`list`

    :return: crsLists is a list of disjoint crsLists.
    :rtype: :py:class:`list`
    """
    crsArray = np.matrix(crsList)
    distances = scipy.spatial.distance.cdist(crsArray, crsArray)

    crsLists = []
    usedIdx = set()
    for startingIndex in range(len(crsList)):
        if not startingIndex in usedIdx:
            currCluster = set([startingIndex])
            newCluster = {index for index, distance in enumerate(distances[startingIndex]) if index not in currCluster and distance <= dcutoff}
            currCluster.update(newCluster)
            while len(newCluster):
                newCluster = {index for oldIndex in newCluster for index, distance in enumerate(distances[oldIndex]) if index not in currCluster and distance <= dcutoff}
                currCluster.update(newCluster)

            usedIdx.update(currCluster)
            crsLists.append([crsList[index] for index in currCluster])
    return crsLists

import itertools
def createSymmetryAtoms(atomList, rotationMats, orthoMat, xs, ys, zs):
    """Creates and returns a list of all symmetry atoms.

    :param atomList:
    :type atomList: :py:class:`list`
    :param rotationMats:
    :type rotationMats: :py:class:`list`
    :param orthoMat:
    :type orthoMat: :py:class:`list`
    :param xs:
    :type xs: :py:class:`list`
    :param ys:
    :type ys: :py:class:`list`
    :param zs:
    :type zs: py:class:`list`

    :return: allAtoms
    :rtype: :py:class:`list`
    """
    allAtoms = []
    for symmetry in itertools.product([-1, 0, 1],[-1, 0, 1],[-1, 0, 1],range(len(rotationMats))):
        if symmetry == (0,0,0,0):
            allAtoms.extend([SymAtom(atom, atom.coord, symmetry) for atom in atomList])
        else:
            rMat = rotationMats[symmetry[3]]
            otMat = np.dot(orthoMat, symmetry[0:3])
            coordList = [np.dot(rMat[:, 0:3], atom.coord) + rMat[:, 3] + otMat for atom in atomList]
            allAtoms.extend([SymAtom(atom, coord, symmetry) for atom,coord in zip(atomList,coordList)
                if xs[0] - 5 <= coord[0] <= xs[-1] + 5 and ys[0] - 5 <= coord[1] <= ys[-1] + 5 and zs[0] - 5 <= coord[2] <= zs[-1] + 5])

    return allAtoms

class SymAtom:
    """A wrapper class to the `BioPDB.atom` class, delegating all BioPDB atom class methods and data members except having its own symmetry and coordination."""

    def __init__(self, atom, coord, symmetry):
        """`pdb_eda.densityAnalysis.symAtom` initializer.

        :param atom: atom object.
        :type atom: :class:`Bio.PDB.atom`
        :param coord: x,y,z coordinates.
        :type coord: :py:class:`list`
        :param symmetry: i,j,k,r symmetry
        :type symmetry: :py:class:`list`
        """
        self.atom = atom
        self.coord = coord
        self.symmetry = symmetry

    def __getattr__(self, attr):
        return getattr(self.atom, attr)

def getPointDensityFromCrs(densityMatrix, crsCoord):
    """Returns the density of a point.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param crsCoord: crs coordinates.
    :type: :py:class:`list`, :py:class:`set`

    :return: density
    :rtype: :py:class:`float`
    """
    crsCoord = list(crsCoord)
    header = densityMatrix.header
    for ind in range(3):
        if crsCoord[ind] < 0 or crsCoord[ind] >= header.ncrs[ind]:
            crsCoord[ind] -= int(np.floor(crsCoord[ind] / header.crsInterval[ind]) * header.crsInterval[ind])

        if (header.ncrs[ind] <= crsCoord[ind] < header.crsInterval[ind]) or crsCoord[ind] < 0:
            return 0

    return densityMatrix.density[crsCoord[2], crsCoord[1], crsCoord[0]]

def testValidCrs(densityMatrix, crsCoord):
    """Tests whether the crs coordinate is valid.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param crsCoord: crs coordinates.
    :type crsCoord: :py:class:`list`

    :return: bool
    :rtype: :py:class:`bool`
    """
    crsCoord = list(crsCoord)
    header = densityMatrix.header
    for ind in range(3):
        if crsCoord[ind] < 0 or crsCoord[ind] >= header.ncrs[ind]:
            crsCoord[ind] -= int(np.floor(crsCoord[ind] / header.crsInterval[ind]) * header.crsInterval[ind])

        if (header.ncrs[ind] <= crsCoord[ind] < header.crsInterval[ind]) or crsCoord[ind] < 0:
            return False

    return True

def testValidCrsList(densityMatrix, crsList):
    """Tests whether all of the crs coordinates in the list are valid.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param crsList: list of crs coordinates.
    :type: :py:class:`list`, :py:class:`set`

    :return: bool
    :rtype: :py:class:`bool`
    """
    return not any(not testValidCrs(densityMatrix,crs) for crs in crsList)

def createFullCrsList(densityMatrix, cutoff):
    """Returns full crs list for the density matrix.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param cutoff:
    :type cutoff: :py:class:`float`

    :return: crsList
    :rtype: :py:class:`list`
    """
    ## only explore the non-repeating part (<= # xyz intervals) of the density map for blobs
    ncrs = densityMatrix.header.uniqueNcrs
    if cutoff > 0:
        return [ crs for crs in itertools.product(range(ncrs[0]),range(ncrs[1]),range(ncrs[2])) if getPointDensityFromCrs(densityMatrix, crs) >= cutoff ]
    elif cutoff < 0:
        return [ crs for crs in itertools.product(range(ncrs[0]),range(ncrs[1]),range(ncrs[2])) if getPointDensityFromCrs(densityMatrix, crs) <= cutoff ]
    else:
        return None

def _testXyzWithinDistance(xyzCoord1, xyzCoord2, distance):
    """Tests whether two xyzCoords are within a certain distance.

    :param xyzCoord1:
    :type xyzCoord1: :py:class:`list`
    :param xyzCoord2:
    :type xyzCoord2: :py:class:`list`
    :param distance:
    :type distance: :py:class:`float`

    :return: bool
    :rtype: :py:class:`bool`
    """
    return np.sqrt((xyzCoord2[0] - xyzCoord1[0])**2 + (xyzCoord2[1] - xyzCoord1[1])**2 + (xyzCoord2[2] - xyzCoord1[2])**2) <= distance

def getSphereCrsFromXyz(densityMatrix, xyzCoord, radius, densityCutoff=0):
    """Calculate a list of crs coordinates that within a given distance of a xyz point.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param xyzCoord: xyz coordinates.
    :type xyzCoord: :py:class:`list`
    :param radius:
    :type radius: :py:class:`float`
    :param densityCutoff: a density cutoff for all the points wants to be included, defaults to 0
            Default 0 means include every point within the radius.
            If cutoff < 0, include only points with density < cutoff.
            If cutoff > 0, include only points with density > cutoff.
    :type densityCutoff: :py:class:`float`

    :return: crsCoordList of crs coordinates
    :rtype: :py:class:`list`
    """
    crsCoord = densityMatrix.header.xyz2crsCoord(xyzCoord)
    crsRadius = densityMatrix.header.xyz2crsCoord(densityMatrix.origin + [radius, radius, radius])
    crsCoordList = []
    for crs in itertools.product(range(crsCoord[0] - crsRadius[0]-1, crsCoord[0] + crsRadius[0]+1),
                                 range(crsCoord[1] - crsRadius[1]-1, crsCoord[1] + crsRadius[1]+1),
                                 range(crsCoord[2] - crsRadius[2]-1, crsCoord[2] + crsRadius[2]+1)):
        density = getPointDensityFromCrs(densityMatrix,crs)
        if ((0 < densityCutoff < density) or (density < densityCutoff < 0) or densityCutoff == 0) and _testXyzWithinDistance(xyzCoord,densityMatrix.header.crs2xyzCoord(crs),radius):
            crsCoordList.append(crs)

    return crsCoordList

def getSphereCrsFromXyzList(densityMatrix, xyzCoordList, radius, densityCutoff=0):
    """Calculates a list of crs coordinates that within a given distance from a list of xyz points.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param xyzCoordList: xyz coordinates.
    :type xyzCoordList: :py:class:`list`
    :param radius:
    :type radius: :py:class:`float`
    :param densityCutoff: a density cutoff for all the points wants to be included., defaults to 0
            Default 0 means include every point within the radius.
            If cutoff < 0, include only points with density < cutoff.
            If cutoff > 0, include only points with density > cutoff.
    :type densityCutoff: :py:class:`float`

    :return: crsCoordList of crs coordinates
    :rtype: :py:class:`set`
    """
    return {tuple(crsCoord) for xyzCoord in xyzCoordList for crsCoord in getSphereCrsFromXyz(densityMatrix, xyzCoord, radius, densityCutoff)}

def testValidXyz(densityMatrix, xyzCoord, radius):
    """Tests whether all crs coordinates within a given distance of a xyzCoord is within the densityMatrix.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param xyzCoord: xyz coordinates.
    :type xyzCoord: :py:class:`list`
    :param radius:
    :type radius: :py:class:`float`

    :return: bool
    :rtype: :py:class:`bool`
    """
    crsCoord = densityMatrix.header.xyz2crsCoord(xyzCoord)
    crsRadius = densityMatrix.header.xyz2crsCoord(densityMatrix.origin + [radius, radius, radius])
    return not any(not testValidCrs(densityMatrix, crs)
                   for crs in itertools.product(range(crsCoord[0] - crsRadius[0]-1, crsCoord[0] + crsRadius[0]+1),
                                                range(crsCoord[1] - crsRadius[1]-1, crsCoord[1] + crsRadius[1]+1),
                                                range(crsCoord[2] - crsRadius[2]-1, crsCoord[2] + crsRadius[2]+1))
                   if _testXyzWithinDistance(xyzCoord, densityMatrix.header.crs2xyzCoord(crs), radius))

def testValidXyzList(densityMatrix, xyzCoordList, radius):
    """Tests whether all crs coordinates within a given distance of a set of xyzCoords is within the densityMatrix.

    :param densityMatrix:
    :type densityMatrix: :class:`pdb_eda.ccp4.DensityMatrix`
    :param xyzCoordList: list of xyz coordinates.
    :type xyzCoordList: :py:class:`list`
    :param radius:
    :type radius: :py:class:`float`

    :return: bool
    :rtype: :py:class:`bool`
    """
    return not any(not testValidXyz(densityMatrix,xyzCoord,radius) for xyzCoord in xyzCoordList)


# Compile using:
# $ python3 cutils_setup.py build_ext --inplace
# $ cp build/lib.linux-x86_64-3.6/pdb_eda/cutils.cpython-36m-x86_64-linux-gnu.so cutils.so

from setuptools import setup
from Cython.Build import cythonize

setup(
    ext_modules = cythonize("cutils.pyx")
)



from pdb_eda import ccp4
import numpy as np
from pytest import approx

pdbid = '1cbs_diff'  # Fo - Fc
densityObj = ccp4.readFromPDBID(pdbid)
posGreen = [6.1, 21.3, 20.7]  # green, big positive
posRed = [34.3, 25.3, 22.8]  # red, big negative


def test_working():
    assert 1 + 2 == 3.0
    assert [1, 2, 3] == [1, 2, 3]
    assert [1, 2, 3] != [3, 2, 1]


def test_origin_match_LiteMol():
    """Test if the origin is the same as LiteMol calculated."""
    alpha = np.pi / 180 * densityObj.header.alpha
    beta = np.pi / 180 * densityObj.header.beta
    gamma = np.pi / 180 * densityObj.header.gamma

    xscale = densityObj.header.gridLength[0]
    yscale = densityObj.header.gridLength[1]
    zscale = densityObj.header.gridLength[2]

    z1 = np.cos(beta)
    z2 = (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma)
    z3 = np.sqrt(1.0 - z1 * z1 - z2 * z2)

    xAxis = [xscale, 0.0, 0.0]
    yAxis = [np.cos(gamma) * yscale, np.sin(gamma) * yscale, 0.0]
    zAxis = [z1 * zscale, z2 * zscale, z3 * zscale]

    if densityObj.header.futureUse[-3] == 0.0 and densityObj.header.futureUse[-2] == 0.0 and densityObj.header.futureUse[-1] == 0.0:
        origin = [
            xAxis[0] * densityObj.header.crsStart[densityObj.header.map2xyz[0]] + yAxis[0] * densityObj.header.crsStart[densityObj.header.map2xyz[1]] +
            zAxis[0] * densityObj.header.crsStart[densityObj.header.map2xyz[2]],
            yAxis[1] * densityObj.header.crsStart[densityObj.header.map2xyz[1]] + zAxis[1] * densityObj.header.crsStart[densityObj.header.map2xyz[2]],
            zAxis[2] * densityObj.header.crsStart[densityObj.header.map2xyz[2]]
        ]
    else:
        origin = [densityObj.header.originEM[0], densityObj.header.originEM[1], densityObj.header.originEM[2]]
        # [-6.0065791481419613, -1.783500051498413, -1.7638636502352627]
        # array([-6.00657915, -1.78350005, -1.76386365])

    for x, y in zip(origin, densityObj.origin):
        assert x == approx(y)


def test_xyz_crs_conversion():
    """Test that converting from crs to xyz and back to crs, or converting from xyz to crs back to xyz will give be the same as originals"""
    assert [60, 60, 60] == approx(densityObj.header.xyz2crsCoord(densityObj.header.crs2xyzCoord([60, 60, 60])))
    assert [80, 60, 60] == approx(densityObj.header.xyz2crsCoord(densityObj.header.crs2xyzCoord([80, 60, 60])))  # column out of boundary
    assert [60, 160, 60] == approx(densityObj.header.xyz2crsCoord(densityObj.header.crs2xyzCoord([60, 160, 60])))  # Row out of boundary
    assert [60, 60, 160] == approx(densityObj.header.xyz2crsCoord(densityObj.header.crs2xyzCoord([60, 60, 160])))  # Section out of boundary

    assert posRed == approx(densityObj.header.crs2xyzCoord(densityObj.header.xyz2crsCoord(posRed)), abs=np.max(densityObj.header.gridLength)/2) # [46, 67, 42]
    assert posGreen == approx(densityObj.header.crs2xyzCoord(densityObj.header.xyz2crsCoord(posGreen)), abs=np.max(densityObj.header.gridLength)/2) # [39, 20, 38]


def test_aberrant_point():
    """Test red and green are both beyond 3 standard deviations of overall all density"""
    assert densityObj.getPointDensityFromXyz(posRed) < - 3 * densityObj.header.rmsd
    assert densityObj.getPointDensityFromXyz(posGreen) > 3 * densityObj.header.rmsd


def test_crs_edge_cases():
    """Test some edge cases of crs coordinates, and check if they were handled correctly."""
    assert densityObj.getPointDensityFromCrs([8, 1, 30]) == densityObj.getPointDensityFromCrs([8, 77, 30])  # Data repeating after given interval number
    assert densityObj.getPointDensityFromCrs([88, 77, 30]) == densityObj.getPointDensityFromCrs([8, 77, 30])  # Out of crs boundary, so bring it back n interval number
    assert densityObj.getPointDensityFromCrs([78, 77, 30]) == 0  # If a crs coordinate was not provided in the data, set it to 0


def test_aberrant_blob():
    """Test the findAberrantBlobs funtion"""
    densityObj.density[:20, :20, :20] = 0
    centerXYZ = densityObj.header.crs2xyzCoord([10, 10, 10])
    densityCutoff = 3 * densityObj.header.rmsd

    # Set up green blobs
    densityObj.density[11:13, 11:13, 11:13] = 1  # (x,y,z), Volume 2*2*2=8
    densityObj.density[7:10, 11:14, 7:10] = 1  # (-x,y,-z), Volume 3:3:3=27

    # Set up red blobs
    densityObj.density[11:13, 7:9, 7:9] = -1  # (-x,-y,z), Volume 2*2*2=8
    densityObj.density[7:10, 7:10, 11:14] = -1  # (x,-y,-z), Volume 3*3*3=27

    calcGreen = densityObj.findAberrantBlobs(centerXYZ, 5, densityCutoff)
    centroid1 = [(densityObj.header.crs2xyzCoord([11, 11, 11])[i] + densityObj.header.crs2xyzCoord([12, 12, 12])[i])/2 for i in range(3)]
    trueGreen = [ccp4.DensityBlob(centroid1, 8, densityObj.header.unitVolume * 8, [], densityObj.header),
                 ccp4.DensityBlob(densityObj.header.crs2xyzCoord([8, 12, 8]), 27, [], densityObj.header.unitVolume * 27, densityObj.header)]

    calcGreen.sort(key=lambda x: (x.volume, x.totalDensity))
    trueGreen.sort(key=lambda x: (x.volume, x.totalDensity))
    for i in range(0, len(calcGreen)):
        assert calcGreen[i] == trueGreen[i]

    calcRed = densityObj.findAberrantBlobs(centerXYZ, 5, -1 * densityCutoff)
    centroid2 = [(densityObj.header.crs2xyzCoord([7, 7, 11])[i] + densityObj.header.crs2xyzCoord([8, 8, 12])[i]) / 2 for i in range(3)]
    trueRed = [ccp4.DensityBlob(centroid2, -8, densityObj.header.unitVolume * 8, [], densityObj.header),
               ccp4.DensityBlob(densityObj.header.crs2xyzCoord([12, 8, 8]), -27, [], densityObj.header.unitVolume * 27, densityObj.header)]

    calcRed.sort(key=lambda x: (x.volume, x.totalDensity))
    trueRed.sort(key=lambda x: (x.volume, x.totalDensity))

    for i in range(0, len(calcRed)):
        assert calcRed[i] == trueRed[i]


def test_merge_blob():
    """Test the merge funtion in class DensityBlob"""
    densityObj.density[:20, :20, :20] = 0
    centerXYZ = densityObj.header.crs2xyzCoord([10, 10, 10])
    densityCutoff = 3 * densityObj.header.rmsd

    # Set up one green and one red blob
    densityObj.density[11:13, 11:13, 11:13] = 1
    calc1 = densityObj.findAberrantBlobs(centerXYZ, 5, densityCutoff)

    densityObj.density[11:13, 11:13, 11:13] = 0
    densityObj.density[8:10, 8:10, 8:10] = 1
    densityObj.density[10, 10, 10] = 1
    calc2 = densityObj.findAberrantBlobs(centerXYZ, 5, densityCutoff)
    densityObj.density[11:13, 11:13, 11:13] = 1

    calc1[0].merge(calc2[0], densityObj.density)
    trueMerge = ccp4.DensityBlob(centerXYZ, 17, densityObj.header.unitVolume * 17, [], densityObj.header)

    assert calc1[0].testOverlap(calc2[0])
    assert calc1[0] == trueMerge

"""
def test_aberrant():
    """Non-orthogonal electron density map"""
    id1 = '2IM3'  # 2Fo - Fc
    metal = [217, -54.7, -17]
    # [221.8, -58.1, -14.5]
    #[220.9, -53.6, -10.2]
    #[224.312, -57.515, -13.837]
    obj1 = ccp4.readFromPDBID(id1)

    assert obj1.getPointDensityFromXyz(metal) == approx(0.44502562284469604)
"""



"""
The generate parameters mode command line interface
  Generates initial parameters needed for pdb_eda calculations.

Usage:
    pdb_eda generate -h | --help
    pdb_eda generate atom-type <out-jsonfile> [--residues=<comma-separated-residues>] [--allow-errors] [--default-slope=<default-slope>] [--F000]
    pdb_eda generate prevalence <pdbid-file> <out-jsonfile> [--testing]
    pdb_eda generate parameters <in-atom-types> <in-prevalence-file> <out-params-file> <out-pdbid-file> [--params=<params-file>] [--min-atom-types=<min-atom-types>] [--min-atoms=<min-atoms>] [--max-atoms=<max-atoms>] [--max-resolution=<max-resolution>] [--min-resolution=<min-resolution>]

Options:
    atom-type                               Generates atom types and other parameters from the chemical component descriptions.
    --residues=<comma-separated-residues>   Limit residues and atom-types to the given comma-separated residues. [default: ]
    --allow-errors                          Allow residues with errors.
    --default-slope=<default-slope>         Default b-factor correction slope to start with. [default: -0.5]
    --F000                                  Generate parameters for F000 calculations.
    prevalence                              Generate prevalence report of residues and atoms across a set of pdbids.
    --testing                               Run only a single process for testing purposes.
    parameters                              Generate parameters file.
    --params=<params-file>                  Overriding parameters file that includes radii, slopes, etc. Useful for including currently optimized atom types. [default: ]
    --min-atom-types=<min-atom-types>       Minimum number of a given atom-type required. [default: 5]
    --min-atoms=<min-atoms>                 Minimum number of parameter atoms required. [default: 500]
    --max-atoms=<max-atoms>                 Maximum number of (all) parameter atoms allowed. Useful to selecting entries that are reasonably fast to analyze [default: 5000]
    --max-resolution=<max-resolution>       Maximum x-ray crystallographic resolution to allow. [default: 3.5]
    --min-resolution=<min-resolution>       Minimum x-ray crystallographic resolution to allow. [default: 0]


Unique atom-types are generated based on a "chemical coloring" approach that uses bond-type, element, estimated electrons, and aromatic information.

The wwPDB chemical components file is located at:
ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif
http://remediation.wwpdb.org/data/ccd

List of all wwPDB entry IDs along with entry type.
ftp://ftp.wwpdb.org/pub/pdb/derived_data/pdb_entry_type.txt
# Select only the protein, diffraction entries
% grep "diffraction" pdb_entry_type.txt | grep "prot" > pdb_prot_diffraction_only.txt

# Faster to download all gzipped compressed PDB entries directly from wwPDB before analyzing
% cd pdb_data
% rsync -rLpt -v -z --port=33444 rsync.rcsb.org::ftp_data/structures/all/pdb/ ./

Amino Acid Residues: ALA,ARG,ASN,ASP,CYS,GLN,GLU,GLY,HIS,ILE,LEU,LYS,MET,PHE,PRO,SER,THR,TRP,TYR,VAL
Nucleic Acid Residues: A,C,G,I,U,DA,DC,DG,DI,DT,DU

Typical order of use:
1) atom-type: generate new set of atom types to optimize.
2) prevalence: analyze a set of PDB entries for atom-type prevalence.
3) parameters: generate new parameters file and list of PDB IDs for optimization.
"""

import CifFile
import docopt
import urllib.request
import os
import collections
import json
import numpy as np
import multiprocessing

from . import __version__
from . import densityAnalysis
from . import fileUtils

componentsFilename = "components.cif"

elementElectrons = {'H': 1, 'HE': 2, 'LI': 3, 'BE': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'NE': 10, 'NA': 11, 'MG': 12, 'AL': 13, 'SI': 14, 'P': 15, 'S': 16, 'CL': 17, 'AR': 18, 'K': 19, 'CA': 20, 'SC': 21, 'TI': 22,
                    'V': 23, 'CR': 24, 'MN': 25, 'FE': 26, 'CO': 27, 'NI': 28, 'CU': 29, 'ZN': 30, 'GA': 31, 'GE': 32, 'AS': 33, 'SE': 34, 'BR': 35, 'RB': 37, 'SR': 38, 'Y': 39, 'ZR': 40, 'NB': 41, 'MO': 42, 'TC': 43,
                    'RU': 44, 'RH': 45, 'PD': 46, 'AG': 47, 'CD': 48, 'IN': 49, 'SN': 50, 'SB': 51, 'TE': 52, 'I': 53, 'CS': 55, 'BA': 56, 'LA': 57, 'CE': 58, 'PR': 59, 'ND': 60, 'PM': 61, 'SM': 62, 'EU': 63, 'GD': 64,
                    'TB': 65, 'DY': 66, 'HO': 67, 'ER': 68, 'TM': 69, 'YB': 70, 'LU': 71, 'HF': 72, 'TA': 73, 'W': 74, 'RE': 75, 'OS': 76, 'IR': 77, 'PT': 78, 'AU': 79, 'HG': 80, 'TL': 81, 'PB': 82, 'BI': 83, 'PO': 84,
                    'RA': 88, 'AC': 89, 'TH': 90, 'PA': 91, 'U': 92, 'NP': 93, 'PU': 94, 'AM': 95}

elementAtomicRadii = {'H': 0.25, 'HE': 1.2, 'LI': 1.45, 'BE': 1.05, 'B': 0.85, 'C': 0.7, 'N': 0.65, 'O': 0.6, 'F': 0.5, 'NE': 1.6, 'NA': 1.8, 'MG': 1.5, 'AL': 1.25, 'SI': 1.1, 'P': 1.0, 'S': 1.0, 'CL': 1.0, 'AR': 0.71,
                      'K': 2.2, 'CA': 1.8, 'SC': 1.6, 'TI': 1.4, 'V': 1.35, 'CR': 1.4, 'MN': 1.4, 'FE': 1.4, 'CO': 1.35, 'NI': 1.35, 'CU': 1.35, 'ZN': 1.35, 'GA': 1.3, 'GE': 1.25, 'AS': 1.15, 'SE': 1.15, 'BR': 1.15,
                      'RB': 2.35, 'SR': 2.0, 'Y': 1.8, 'ZR': 1.55, 'NB': 1.45, 'MO': 1.45, 'TC': 1.35, 'RU': 1.3, 'RH': 1.35, 'PD': 1.4, 'AG': 1.6, 'CD': 1.55, 'IN': 1.55, 'SN': 1.45, 'SB': 1.45, 'TE': 1.4, 'I': 1.4,
                      'CS': 2.6, 'BA': 2.15, 'LA': 1.95, 'CE': 1.85, 'PR': 1.85, 'ND': 1.85, 'PM': 1.85, 'SM': 1.85, 'EU': 1.85, 'GD': 1.8, 'TB': 1.75, 'DY': 1.75, 'HO': 1.75, 'ER': 1.75, 'TM': 1.75, 'YB': 1.75,
                      'LU': 1.75, 'HF': 1.55, 'TA': 1.45, 'W': 1.35, 'RE': 1.35, 'OS': 1.3, 'IR': 1.35, 'PT': 1.35, 'AU': 1.35, 'HG': 1.5, 'TL': 1.9, 'PB': 1.8, 'BI': 1.6, 'PO': 1.9, 'RA': 2.15, 'AC': 1.95, 'TH': 1.8,
                      'PA': 1.8, 'U': 1.75, 'NP': 1.75, 'PU': 1.75, 'AM': 1.75}

oxygenDoubleBondColor = "O.N.8.DOUB"
oxygenSingleBondColor = "O.N.9.SING"

def main():
    args = docopt.docopt(__doc__, version=__version__)
    if args["--help"]:
        print(__doc__)
        exit(0)

    if args["atom-type"]:
        if os.path.isfile("components_info.json"):
            with open("components_info.json", 'r') as jsonFile:
                componentsInfo = json.load(jsonFile)
        else:
            componentsInfo = processComponents()

        if not os.path.isfile("components_info.json"):
            with open("components_info.json", 'w') as outFile:
                print(json.dumps(componentsInfo, indent=2, sort_keys=True), file=outFile)

        componentsInfo["errors"] = set(componentsInfo["errors"])
        allowedResidueTypes = set(args["--residues"].split(",")) if args["--residues"] else set()
        args["--default-slope"] = float(args["--default-slope"])

        if args["--F000"]:
            initialParams = {"full_atom_name_map_electrons": {}, "element_map_electrons": elementElectrons }
        else:
            initialParams = { "full_atom_name_map_atom_type" : {}, "full_atom_name_map_electrons" : {}, "leaving_atoms" : [], "radii" : {}, "slopes" : {}, "bonded_atoms" : {} }

        fullAtomName2AtomType = {}

        for residue in componentsInfo["residues"].values():
            residue["estimated_electrons"] = 0
            for atom in residue["atoms"].values():
                atom["num_bound_hydrogens"] = sum(1 for atomName,bondType,aromatic,stereo in atom["bonds"] if atomName in residue["atoms"] and residue["atoms"][atomName]["element"] == "H" and residue["atoms"][atomName]["leaving"] == atom["leaving"])
                if atom["element"] in elementElectrons:
                    try:
                        atom["charge"] = float(atom["charge"])
                    except:
                        atom["charge"] = 0
                    atom["estimated_electrons"] = elementElectrons[atom["element"]] + atom["num_bound_hydrogens"] - atom["charge"]
                else:
                    atom["estimated_electrons"] = 0
                    componentsInfo["errors"].add(residue["name"])

            for atom in residue["atoms"].values():
                atom["element_color"] = atom["element"] + "." + atom["aromatic"] + "." + str(int(atom["estimated_electrons"]))
            for atom in residue["atoms"].values():
                atom["element_bond_colors"] = [ residue["atoms"][atomName]["element_color"] + "." +  bondTyping(bondType,aromatic) for atomName,bondType,aromatic,stereo in atom["bonds"]
                                                if atomName in residue["atoms"] and (atom["leaving"] == "Y" or atom["leaving"] == residue["atoms"][atomName]["leaving"]) ]
                atom["full_element_color"] = atom["element_color"] + "#" + "_".join(sorted(atom["element_bond_colors"]))

            # identify and adjust oxygen atoms in resonance.
            for testAtom in residue["atoms"].values():
                if oxygenDoubleBondColor in testAtom["element_bond_colors"] and oxygenSingleBondColor in testAtom["element_bond_colors"]:
                    oxygenAtomTuples = [(residue["atoms"][atomName],residue["atoms"][atomName]["element_color"] + "." + bondTyping(bondType,aromatic)) for atomName,bondType,aromatic,stereo in testAtom["bonds"]
                                        if atomName in residue["atoms"] and residue["atoms"][atomName]["element"] == "O"  and testAtom["leaving"] == "Y" or testAtom["leaving"] == residue["atoms"][atomName]["leaving"] ]
                    resonanceOxygenAtoms = [atom for atom,bondColor in oxygenAtomTuples if bondColor == oxygenDoubleBondColor or bondColor == oxygenSingleBondColor ]
                    if len(set(atom["estimated_electrons"] for atom in resonanceOxygenAtoms)) > 1:
                        averageElectrons = np.mean([atom["estimated_electrons"] for atom in resonanceOxygenAtoms])
                        longestFullElementColor = ""
                        for atom in resonanceOxygenAtoms:
                            atom["estimated_electrons"] = averageElectrons
                            atom["element_color"] = atom["element"] + "." + atom["aromatic"] + "." + str(float(atom["estimated_electrons"]))[:5]
                            atom["element_bond_colors"] = [ residue["atoms"][atomName]["element_color"] + "." +  "RESON" for atomName,bondType,aromatic,stereo in atom["bonds"]
                                                            if atomName in residue["atoms"]  and (atom["leaving"] == "Y" or atom["leaving"] == residue["atoms"][atomName]["leaving"]) ]
                            atom["full_element_color"] = atom["element_color"] + "#" + "_".join(sorted(atom["element_bond_colors"]))
                            if len(atom["full_element_color"]) > len(longestFullElementColor):
                                longestFullElementColor = atom["full_element_color"]
                        for atom in resonanceOxygenAtoms:
                            atom["full_element_color"] = longestFullElementColor

            if args["--allow-errors"] or residue["name"] not in componentsInfo["errors"]:
                for atom in residue["atoms"].values():
                    residue["estimated_electrons"] += atom["estimated_electrons"] if atom["leaving"] != "Y" else 0
                residue["estimated_electrons"] = float(np.round(residue["estimated_electrons"]))

                if not allowedResidueTypes or residue["name"] in allowedResidueTypes: # only add certain residues
                    for atom in residue["atoms"].values():
                        if atom["element"] != "H":
                            fullAtomName = residue["name"].strip() + "_" + atom["name"]
                            initialParams["full_atom_name_map_electrons"][fullAtomName] = atom["estimated_electrons"]
                            fullAtomName2AtomType[fullAtomName] = atom["full_element_color"]
                            if not args["--F000"]:
                                initialParams["full_atom_name_map_atom_type"][fullAtomName] = atom["full_element_color"]
                                initialParams["radii"][atom["full_element_color"]] = elementAtomicRadii[atom["element"]]
                                initialParams["slopes"][atom["full_element_color"]] = args["--default-slope"]
                                initialParams["bonded_atoms"][fullAtomName] = [residue["name"].strip() + "_" + atomName for atomName, bondType, aromatic, stereo in atom["bonds"] if residue["atoms"][atomName]["element"] != "H"]
                                if atom["leaving"] == "Y":
                                    initialParams["leaving_atoms"].append(fullAtomName)

        # Print report
        print("Unique Residue Types:",len(set(name.split("_")[0] for name in fullAtomName2AtomType.keys())))
        print("Unique Full Atom Names:",len(set(fullAtomName2AtomType.keys())))
        print("Unique Atom Types:",len(set(fullAtomName2AtomType.values())))
        # testElectrons = collections.defaultdict(set)
        # for name,value in initialParams["full_atom_name_map_atom_type"].items():
        #     testElectrons[value].add(initialParams["full_atom_name_map_electrons"][name])
        # for atomType,values in testElectrons.items():
        #     if len(values) > 1:
        #         print(atomType,values)

        with open(args["<out-jsonfile>"], 'w') if args["<out-jsonfile>"] != "-" else sys.stdout as outFile:
            print(json.dumps(initialParams, indent=2, sort_keys=True), file=outFile)
    elif args["prevalence"]:
        try:
            pdbids = []
            with open(args["<pdbid-file>"], "r") as textFile:
                for pdbid in textFile:
                    pdbids.append(pdbid[0:4])
        except:
            RuntimeError(str("Error: PDB IDs file \"") + args["<pdbid-file>"] + "\" does not exist or is not parsable.")


        if args["--testing"]:
            results = [ processFunction(pdbid) for pdbid in pdbids ]
        else:
            with multiprocessing.Pool() as pool:
                results = pool.map(processFunction, pdbids)

        pdbidInfo = {}
        for resultFilename in results:
            if resultFilename:
                try:
                    with open(resultFilename, 'r') as jsonFile:
                        result = json.load(jsonFile)
                        pdbidInfo[result["pdbid"]] = result
                    os.remove(resultFilename)
                except:
                    pass

        totalResidueAtoms = collections.defaultdict(int)
        totalElements = collections.defaultdict(int)
        totalResidues = collections.defaultdict(int)
        for pdbid,info in pdbidInfo.items():
            for name,count in info["full_atom_name_counts"].items():
                totalResidueAtoms[name] += count
            for name,count in info["element_counts"].items():
                totalElements[name] += count
            for name,count in info["residue_counts"].items():
                totalResidues[name] += count

        with open(args["<out-jsonfile>"], 'w') if args["<out-jsonfile>"] != "-" else sys.stdout as outFile:
            jsonOutput = { "pdbid_info" : pdbidInfo, "full_atom_name_counts" : totalResidueAtoms, "residue_counts" : totalResidues, "element_counts" : totalElements }
            print(json.dumps(jsonOutput, indent=2, sort_keys=True), file=outFile)
    elif args["parameters"]:
        args["--max-resolution"] = float(args["--max-resolution"])
        args["--min-resolution"] = float(args["--min-resolution"])
        args["--min-atom-types"] = int(args["--min-atom-types"])
        args["--min-atoms"] = int(args["--min-atoms"])
        args["--max-atoms"] = int(args["--max-atoms"])

        if args["--params"]:
            try:
                with open(args["--params"], 'r') as paramsFile:
                    overrideParams = json.load(paramsFile)
            except:
                RuntimeError(str("Error: params file \"") + args["--params"] + "\" does not exist or is not parsable.")
        else:
            overrideParams = None

        try:
            with open(args["<in-atom-types>"], 'r') as paramsFile:
                initialParams = json.load(paramsFile)
        except:
            RuntimeError(str("Error: params file \"") + args["<in-atom-types>"] + "\" does not exist or is not parsable.")

        try:
            with open(args["<in-prevalence-file>"], 'r') as jsonFile:
                prevalenceInfo = json.load(jsonFile)
        except:
            RuntimeError(str("Error: prevalence file \"") + args["<in-prevalence-file>"] + "\" does not exist or is not parsable.")

        currentPDBinfo = { pdbid:info for pdbid,info in prevalenceInfo["pdbid_info"].items()
                           if info["properties"]["resolution"] >= args["--min-resolution"] and info["properties"]["resolution"] <= args["--max-resolution"] }

        testingFullAtomNames = [fullAtomName for fullAtomName in initialParams["full_atom_name_map_atom_type"].keys()
                                if fullAtomName not in initialParams["leaving_atoms"] and (not overrideParams or fullAtomName not in overrideParams["full_atom_name_map_atom_type"])]
        testingAtomTypes = set(initialParams["full_atom_name_map_atom_type"][fullAtomName] for fullAtomName in testingFullAtomNames)

        allFullAtomNames = list(testingFullAtomNames)
        if overrideParams:
            allFullAtomNames.extend([fullAtomName for fullAtomName in overrideParams["full_atom_name_map_atom_type"].keys() if fullAtomName not in overrideParams["leaving_atoms"]])

        pdbids = []
        for pdbid, info in currentPDBinfo.items():
            atomTypeSum = { atomType:0 for atomType in testingAtomTypes }
            for fullAtomName in testingFullAtomNames:
                atomTypeSum[initialParams["full_atom_name_map_atom_type"][fullAtomName]] += info["full_atom_name_counts"][fullAtomName] if fullAtomName in info["full_atom_name_counts"] else 0

            analyzableAtoms = sum(atomTypeSum.values())
            totalAtoms = sum(info["full_atom_name_counts"][fullAtomName] for fullAtomName in allFullAtomNames if fullAtomName in info["full_atom_name_counts"])
            if all(count >= args["--min-atom-types"] for count in atomTypeSum.values()) and analyzableAtoms >= args["--min-atoms"] and totalAtoms <= args["--max-atoms"]:
                pdbids.append(pdbid)

        with open(args['<out-pdbid-file>'], "w") if args["<out-pdbid-file>"] != "-" else sys.stdout as txtFile:
            print("\n".join(pdbids), file=txtFile)

        if overrideParams:
            initialParams["full_atom_name_map_atom_type"].update(overrideParams["full_atom_name_map_atom_type"])
            initialParams["full_atom_name_map_electrons"].update(overrideParams["full_atom_name_map_electrons"])
            initialParams["radii"].update(overrideParams["radii"])
            initialParams["slopes"].update(overrideParams["slopes"])
            leavingAtoms = set(initialParams["leaving_atoms"])
            leavingAtoms.update(overrideParams["leaving_atoms"])
            initialParams["leaving_atoms"] = list(leavingAtoms)
            initialParams["optimize"] = [atomType for atomType in initialParams["radii"].keys() if atomType not in overrideParams["radii"]]

        with open(args["<out-params-file>"], 'w') if args["<out-params-file>"] != "-" else sys.stdout as outFile:
            print(json.dumps(initialParams, indent=2, sort_keys=True), file=outFile)


def processComponents():
    """Retrieves and extracts residues, atoms, and bonds from the wwPDB chemical components Cif file.

    :return: component information
    :rtype: :py:class:`dict`
    """
    if not os.path.isfile(componentsFilename):
        urllib.request.urlretrieve("ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif", componentsFilename)

    print("Parsing the components.cif file.  This takes a few minutes.")
    components = CifFile.ReadCif(componentsFilename)

    print("Processing the components.")
    residues = {}
    errors = set()
    for residueName in components.keys():
        residue = components[residueName]
        residueName = residueName.upper().strip()
        if "_chem_comp_atom.atom_id" in residue and "_chem_comp_atom.charge" in residue and "_chem_comp_atom.type_symbol" in residue and "_chem_comp_atom.pdbx_leaving_atom_flag" in residue and \
                "_chem_comp_atom.pdbx_aromatic_flag" in residue:
            atoms = { name : {"name": name, "charge": charge, "element" : element, "leaving" : leaving, "aromatic" : aromatic, "bonds" : [] } for (name, charge, element, leaving, aromatic) in \
                      zip(residue["_chem_comp_atom.atom_id"], residue["_chem_comp_atom.charge"], residue["_chem_comp_atom.type_symbol"], residue["_chem_comp_atom.pdbx_leaving_atom_flag"], residue["_chem_comp_atom.pdbx_aromatic_flag"]) }
            if "_chem_comp_bond.atom_id_1" in residue and "_chem_comp_bond.atom_id_2" in residue and "_chem_comp_bond.value_order" in residue and \
                "_chem_comp_bond.pdbx_aromatic_flag" in residue and "_chem_comp_bond.pdbx_stereo_config" in residue:
                for (atomName1, atomName2, bondType,isAromatic,stereo) in \
                    zip(residue["_chem_comp_bond.atom_id_1"],residue["_chem_comp_bond.atom_id_2"],residue["_chem_comp_bond.value_order"], residue["_chem_comp_bond.pdbx_aromatic_flag"],residue["_chem_comp_bond.pdbx_stereo_config"]):
                    if atomName1 in atoms:
                        atoms[atomName1]["bonds"].append((atomName2,bondType,isAromatic,stereo))
                    else:
                        errors.add(residueName)
                    if atomName2 in atoms:
                        atoms[atomName2]["bonds"].append((atomName1,bondType,isAromatic,stereo))
                    else:
                        errors.add(residueName)
            residues[residueName] = { "name" : residueName, "atoms" : atoms }

    return { "residues" : residues, "errors" : list(errors) }


def bondTyping(bondType, aromatic):
    """Returns type of bond based on given bondType and whether it was marked aromatic.

    :param bondType:
    :type bondType: :py:class:`str`
    :param aromatic:
    :type aromatic: :py:class:`str`

    :return: bondType
    :rtype: :py:class:`str`
    """
    return bondType if aromatic == "N" else "AROM"

def processFunction(pdbid):
    """Process function for multiprocessing to analyze a single pdb entry.

    :param pdbid: pdbid for entry to download and analyze.
    :type pdbid: :py:class:`str`

    :return: resultFilename
    :rtype: :py:class:`str`
    """
    if not densityAnalysis.testCCP4URL(pdbid):
        return 0

    analyzer = densityAnalysis.fromPDBid(pdbid, ccp4density=False, ccp4diff=False)
    if not analyzer:
        return 0

    info = {}
    info["pdbid"] = pdbid
    info["properties"] = {property: value for (property, value) in analyzer.biopdbObj.header.items()}
    info["properties"]["resolution"] = float(analyzer.pdbObj.header.resolution)
    info["properties"]['space_group'] = analyzer.pdbObj.header.spaceGroup
    info["full_atom_name_counts"] = collections.Counter(densityAnalysis.residueAtomName(atom) for residue in analyzer.biopdbObj.get_residues() for atom in residue.get_atoms())
    info["element_counts"] = collections.Counter(atom.element for residue in analyzer.biopdbObj.get_residues() for atom in residue.get_atoms())
    info["residue_counts"] = collections.Counter(residue.resname for residue in analyzer.biopdbObj.get_residues())

    resultFilename = fileUtils.createTempJSONFile(info, "tempResults_")
    return resultFilename



"""
PDB Electron Density Analysis (pdb_eda.densityAnalysis)
-------------------------------------------------------

This module provides methods for the creation of the :class:`pdb_eda.densityAnalysis` class given a PDB id,
along with methods to analyze its 2Fo-Fc and Fo-Fc electron density maps.
"""

import copy
import urllib.request
import os.path
import gzip

import itertools
import collections
import json
import numpy as np
import Bio.PDB as biopdb
import scipy.spatial
from scipy import stats


from . import ccp4
from . import pdbParser

try:
    from . import cutils as utils
except ImportError:
    from . import utils

## Starting data originally from https://arxiv.org/pdf/0804.2488.pdf
paramsPath = os.path.join(os.path.dirname(__file__), 'conf/optimized_params.json')
f000ParamsPath = os.path.join(os.path.dirname(__file__), 'conf/f000_parameters.json.gz')

paramsGlobal = None
with open(paramsPath, 'r') as fh:
    paramsGlobal = json.load(fh)

radiiGlobal = paramsGlobal['radii']
slopesGlobal = paramsGlobal['slopes']
bondedAtomsGlobal = paramsGlobal['bonded_atoms']
elementElectronsGlobal = None
masterFullAtomNameMapElectronsGlobal = None
fullAtomNameMapElectronsGlobal = paramsGlobal['full_atom_name_map_electrons']
fullAtomNameMapAtomTypeGlobal = paramsGlobal['full_atom_name_map_atom_type']
atomTypeLengthGlobal = max(len(atomType) for atomType in fullAtomNameMapAtomTypeGlobal.values()) + 5

def setGlobals(params):
    """Sets global parameters.  Typically used for optimizing parameters.

    :param  params: dictionary of parameters needed for pdb_eda calculations.
    :type params: :py:class:`dict`
    """
    global paramsGlobal
    global radiiGlobal
    global slopesGlobal
    global bondedAtomsGlobal
    global fullAtomNameMapElectronsGlobal
    global fullAtomNameMapAtomTypeGlobal
    global atomTypeLengthGlobal

    paramsGlobal = params
    radiiGlobal = paramsGlobal['radii']
    slopesGlobal = paramsGlobal['slopes']
    bondedAtomsGlobal = paramsGlobal['bonded_atoms']
    fullAtomNameMapElectronsGlobal = paramsGlobal['full_atom_name_map_electrons']
    fullAtomNameMapAtomTypeGlobal = paramsGlobal['full_atom_name_map_atom_type']
    atomTypeLengthGlobal = max(len(atomType) for atomType in fullAtomNameMapAtomTypeGlobal.values()) + 5

def loadF000Parameters():
    """Loads and assigns global parameters needed for F000 estimation."""
    with gzip.open(f000ParamsPath, 'rt') as gzipFile:
        f000Params = json.load(gzipFile)

    global elementElectronsGlobal
    elementElectronsGlobal = f000Params["element_map_electrons"]
    global masterFullAtomNameMapElectronsGlobal
    masterFullAtomNameMapElectronsGlobal = f000Params["full_atom_name_map_electrons"]

ccp4urlPrefix = "http://www.ebi.ac.uk/pdbe/coordinates/files/"
ccp4folder = './ccp4_data/'
pdbfolder = './pdb_data/'
pdburlPrefix = "http://ftp.rcsb.org/pub/pdb/data/structures/all/pdb/"
mmcifurlPrefix = "http://ftp.rcsb.org/pub/pdb/data/structures/all/mmCIF/"


def fromPDBid(pdbid, ccp4density=True, ccp4diff=True, pdbbio=True, pdbi=True, downloadFile=True, mmcif=False):
    """Creates :class:`pdb_eda.densityAnalysis.DensityAnalysis` object given the PDB id if the id is valid
    and the structure has electron density file available.

    :param pdbid: PDB entry ID.
    :type pdbid: :py:class:`str`
    :param ccp4density: Whether to generate ccp4 density object, defaults to :py:obj:`True`.
    :type ccp4density: :py:class:`bool`
    :param ccp4diff: Whether to generate in default of ccp4 difference density object, defaults to :py:obj:`True`.
    :type ccp4diff: :py:class:`bool`
    :param pdbbio: Whether to generate in default of bio.PDB object, defaults to :py:obj:`True`.
    :type pdbbio: :py:class:`bool`
    :param pdbi: Whether to generate in default of PDB object, defaults to :py:obj:`True`.
    :type pdbi: :py:class:`bool`
    :param downloadFile: Whether to save the downloaded ccp4 density, ccp4 difference density, and PDB file, defaults to :py:obj:`True`.
    :type downloadFile: :py:class:`bool`
    :param mmcif: Whether to download the mmCif file, defaults to :py:obj:`False`.
    :type mmcif: :py:class:`bool`

    :return: densityAnalysis object
    :rtype: :class:`pdb_eda.densityAnalysis.DensityAnalysis`
    """
    pdbid = pdbid.lower()

    densityObj = None
    diffDensityObj = None
    pdbObj = None
    biopdbObj = None

    try:
        if ccp4density:
            ## ccp4 2Fo - Fc map parser
            if downloadFile:
                if not os.path.exists(ccp4folder):
                    os.makedirs(ccp4folder)

                ccp4file = ccp4folder + pdbid + '.ccp4'
                if not os.path.isfile(ccp4file):
                    url = ccp4urlPrefix + pdbid + '.ccp4'
                    urllib.request.urlretrieve(url, ccp4file)
                densityObj = ccp4.read(ccp4file, pdbid)
            else:
                densityObj = ccp4.readFromPDBID(pdbid)
            densityObj.densityCutoff = densityObj.meanDensity + 1.5 * densityObj.stdDensity
            densityObj.densityCutoffFromHeader = densityObj.header.densityMean + 1.5 * densityObj.header.rmsd

        if ccp4diff:
            ## ccp4 Fo - Fc map parser
            if downloadFile:
                if not os.path.exists(ccp4folder):
                    os.makedirs(ccp4folder)

                ccp4diffFile = ccp4folder + pdbid + '_diff.ccp4'
                if not os.path.isfile(ccp4diffFile):
                    url = ccp4urlPrefix + pdbid + '_diff.ccp4'
                    urllib.request.urlretrieve(url, ccp4diffFile)

                diffDensityObj = ccp4.read(ccp4diffFile, pdbid)
            else:
                diffDensityObj = ccp4.readFromPDBID(pdbid + '_diff')
            diffDensityObj.diffDensityCutoff = diffDensityObj.meanDensity + 3 * diffDensityObj.stdDensity

        if pdbbio or pdbi:
            pdbfile = pdbfolder + 'pdb' + pdbid + '.ent.gz'
            if not os.path.isfile(pdbfile):
                if not os.path.exists(pdbfolder):
                    os.makedirs(pdbfolder)

                url = pdburlPrefix + 'pdb' + pdbid + '.ent.gz'
                urllib.request.urlretrieve(url, pdbfile)

            if pdbbio:
                # Bio Python PDB parser
                with gzip.open(pdbfile, 'rt') as gzipFile:
                    parser = biopdb.PDBParser(QUIET=True)
                    biopdbObj = parser.get_structure(pdbid, gzipFile)
            if pdbi:
                with gzip.open(pdbfile, 'rt') as gzipFile:
                    pdbObj = pdbParser.readPDBfile(gzipFile)

        if mmcif and downloadFile:
            mmcifFile = pdbfolder + pdbid + '.cif.gz'
            if not os.path.isfile(mmcifFile):
                if not os.path.exists(pdbfolder):
                    os.makedirs(pdbfolder)

                url = mmcifurlPrefix + pdbid + '.cif.gz'
                urllib.request.urlretrieve(url, mmcifFile)
    except:
        return 0

    return DensityAnalysis(pdbid, densityObj, diffDensityObj, biopdbObj, pdbObj)


def fromFile(pdbFile, ccp4DensityFile=None, ccp4DiffDensityFile=None):
    """Creates :class:`pdb_eda.densityAnalysis.DensityAnalysis` object given the appropriate PDB and CCP4 files.

    :param pdbFile: PDB entry file.
    :type pdbFile: :py:class:`str`, :class:`io.IOBase`
    :param ccp4DensityFile: ccp4 density file.
    :type ccp4DensityFile: :py:class:`str`, :class:`io.IOBase`, optional
    :param ccp4DiffDensityFile: ccp4 difference density file.
    :type ccp4DiffDensityFile: :py:class:`str`, :class:`io.IOBase`, optional

    :return: densityAnalysis object
    :rtype: :class:`pdb_eda.densityAnalysis.DensityAnalysis`
    """
    pdbid = "xxxx"
    densityObj = None
    diffDensityObj = None

    try:
        if ccp4DensityFile != None:
            if isinstance(ccp4DensityFile,str):
                densityObj = ccp4.read(ccp4DensityFile, pdbid)
            else:
                densityObj = ccp4.parse(ccp4DensityFile, pdbid)

            densityObj.densityCutoff = densityObj.meanDensity + 1.5 * densityObj.stdDensity
            densityObj.densityCutoffFromHeader = densityObj.header.densityMean + 1.5 * densityObj.header.rmsd

        if ccp4DiffDensityFile != None:
            if isinstance(ccp4DiffDensityFile,str):
                diffDensityObj = ccp4.read(ccp4DiffDensityFile, pdbid)
            else:
                diffDensityObj = ccp4.parse(ccp4DiffDensityFile, pdbid)
            diffDensityObj.diffDensityCutoff = diffDensityObj.meanDensity + 3 * diffDensityObj.stdDensity

        if isinstance(pdbFile,str) and pdbFile.endswith(".gz"):
            with gzip.open(pdbFile, 'rt') as gzipFile:
                parser = biopdb.PDBParser(QUIET=True)
                biopdbObj = parser.get_structure(pdbid, gzipFile)
            with gzip.open(pdbFile, 'rt') as gzipFile:
                pdbObj = pdbParser.readPDBfile(gzipFile)
        else:
            parser = biopdb.PDBParser(QUIET=True)
            biopdbObj = parser.get_structure(pdbid, pdbFile)
            pdbObj = pdbParser.readPDBfile(pdbFile)
    except:
        return 0

    return DensityAnalysis(pdbid, densityObj, diffDensityObj, biopdbObj, pdbObj)


def cleanPDBid(pdbid):
    """Removes PDB entry, ccp4, and mmcif files associated with a PDB id.

    :param pdbid: PDB entry ID.
    :type pdbid: :py:class:`str`

    :return: bool whether the operation was successful.
    :rtype: :py:class:`bool`
    """
    pdbid = pdbid.lower()
    try:
        ccp4file = ccp4folder + pdbid + '.ccp4'
        if os.path.isfile(ccp4file):
            os.remove(ccp4file)

        ccp4diffFile = ccp4folder + pdbid + '_diff.ccp4'
        if os.path.isfile(ccp4diffFile):
            os.remove(ccp4diffFile)

        pdbfile = pdbfolder + 'pdb' + pdbid + '.ent.gz'
        if os.path.isfile(pdbfile):
            os.remove(pdbfile)

        mmcifFile = pdbfolder + pdbid + '.cif.gz'
        if os.path.isfile(mmcifFile):
            os.remove(mmcifFile)
    except:
        return False
    return True

def testCCP4URL(pdbid):
    """Test whether the pdbid has electron density maps by querying if the PDBe API has electron density statistics.
    :param pdbid: PDB entry ID.
    :type pdbid: :py:class:`str`

    :return: bool on test success.
    :rtype: :py:class:`bool`
    """
    try:
        url = "https://www.ebi.ac.uk/pdbe/api/pdb/entry/electron_density_statistics/" + pdbid
        request = urllib.request.urlopen(url)
    except urllib.request.HTTPError:
        return False
    return True


class DensityAnalysis(object):
    """DensityAnalysis class that stores the density, difference density, bio.PDB, and PDB objects."""

    def __init__(self, pdbid, densityObj=None, diffDensityObj=None, biopdbObj=None, pdbObj=None):
        """`densityAnalysis` initializer. Leave `densityObj`, `diffDensityObj`, `biopdbObj` and `pdbObj` as :py:obj:`None`
        to be created. They are not required for initialization but could be required for some methods.

        :param pdbid: PDB entry ID.
        :type pdbid: :py:class:`str`
        :param densityObj: DensityMatrix object.
        :type densityObj: :class:`pdb_eda.ccp4.DensityMatrix`, optional
        :param diffDensityObj: optional DensityMatrix object.
        :type diffDensityObj: :class:`pdb_eda.ccp4.DensityMatrix`, optional
        :param biopdbObj: optional Bio.PDB Structure object.
        :type biopdbObj: :class:`Bio.PDB.Structure.Structure`, optional
        :param pdbObj: optional PDBentry object.
        :type pdbObj: :class:`pdb_eda.pdbParser.PDBentry`, optional
        """
        self.pdbid = pdbid
        self.densityObj = densityObj
        self.diffDensityObj = diffDensityObj
        self.biopdbObj = biopdbObj
        self.pdbObj = pdbObj

        self._symmetryAtoms = None
        self._symmetryOnlyAtoms = None
        self._asymmetryAtoms = None
        self._symmetryAtomCoords = None
        self._symmetryOnlyAtomCoords = None
        self._asymmetryAtomCoords = None
        self._greenBlobList = None
        self._redBlobList = None
        self._blueBlobList = None
        self._fc = None

        self._medians = None
        self._atomCloudDescriptions = None
        self._residueCloudDescriptions = None
        self._domainCloudDescriptions = None
        self._F000 = None
        self._densityElectronRatio = None
        self._numVoxelsAggregated = None
        self._totalAggregatedElectrons = None
        self._totalAggregatedDensity = None

        self._atomTypeOverlapCompleteness = None
        self._atomTypeOverlapIncompleteness = None

    @property
    def symmetryAtoms(self):
        """Returns list of symmetry atoms.

        :return: symmetryAtoms
        :rtype: :py:class:`list`
        """
        if self._symmetryAtoms is None:
            self._calculateSymmetryAtoms()
        return self._symmetryAtoms

    @property
    def symmetryOnlyAtoms(self):
        """Returns list of non-[0,0,0,0] symmetry atoms.

        :return: symmetryAtoms
        :rtype: :py:class:`list`
        """
        if self._symmetryOnlyAtoms is None:
            self._calculateSymmetryAtoms()
        return self._symmetryOnlyAtoms

    @property
    def asymmetryAtoms(self):
        """Returns list of [0,0,0,0] symmetry atoms.

        :return: symmetryAtoms
        :rtype: :py:class:`list`
        """
        if self._asymmetryAtoms is None:
            self._calculateSymmetryAtoms()
        return self._asymmetryAtoms

    @property
    def symmetryAtomCoords(self):
        """Returns list of symmetry atom xyz coordinates.

        :return: symmetryAtomCoords
        :rtype: :py:class:`list`
        """
        if self._symmetryAtoms is None:
            self._calculateSymmetryAtoms()
        return self._symmetryAtomCoords

    @property
    def symmetryOnlyAtomCoords(self):
        """Returns list of non-[0,0,0,0] symmetry atom xyz coordinates.

        :return: symmetryAtomCoords
        :rtype: :py:class:`list`
        """
        if self._symmetryOnlyAtoms is None:
            self._calculateSymmetryAtoms()
        return self._symmetryOnlyAtomCoords

    @property
    def asymmetryAtomCoords(self):
        """Returns list of [0,0,0,0] symmetry atom xyz coordinates.

        :return: symmetryAtomCoords
        :rtype: :py:class:`list`
        """
        if self._asymmetryAtoms is None:
            self._calculateSymmetryAtoms()
        return self._asymmetryAtomCoords

    @property
    def greenBlobList(self):
        """Returns list of all positive significant difference density blobs.

        :return: greenBlobList
        :rtype: :py:class:`list`
        """
        if self._greenBlobList is None:
            self._greenBlobList = self.diffDensityObj.createFullBlobList(self.diffDensityObj.diffDensityCutoff)
        return self._greenBlobList

    @property
    def redBlobList(self):
        """Returns list of all negative significant difference density blobs.

        :return: redBlobList
        :rtype: :py:class:`list`
        """
        if self._redBlobList is None:
            self._redBlobList = self.diffDensityObj.createFullBlobList(-1 * self.diffDensityObj.diffDensityCutoff)
        return self._redBlobList

    @property
    def blueBlobList(self):
        """Returns list of all positive significant density blobs.

        :return: blueBlobList
        :rtype: :py:class:`list`
        """
        if self._blueBlobList is None:
            self._blueBlobList = self.densityObj.createFullBlobList(self.densityObj.densityCutoff)
        return self._blueBlobList

    @property
    def fc(self):
        """Returns the Fc map = 2Fo-Fc - Fo-Fc.

        :return: fc
        :rtype: :class:`pdb_eda.ccp4.DensityMatrix`
        """
        if self._fc is None:
            self._fc = copy.deepcopy(self.densityObj)
            self._fc.density = self.densityObj.density - self.diffDensityObj.density * 2
        return self._fc

    @property
    def fo(self):
        """Returns the Fo map = 2Fo-Fc.

        :return: fo
        :rtype: :class:`pdb_eda.ccp4.DensityMatrix`
        """
        return self.densityObj # using the 2Fo-Fc as the Fo map.

    @property
    def F000(self):
        """Returns estimated F000.  This estimate may not be that accurate.

        :return: F000
        :rtype: :py:class:`float`
        """
        if self._F000 is None:
            self._F000 = self.estimateF000()
        return self._F000

    @property
    def medians(self):
        """Returns median field values calculated per atom type.

        :return: medians
        :rtype: :class:`numpy.array`
        """
        if self._medians is None:
            self.aggregateCloud()
        return self._medians

    @property
    def atomCloudDescriptions(self):
        """Returns aggregated atom cloud descriptions.

        :return: atomCloudDescriptions
        :rtype: :class:`numpy.array`
        """
        if self._atomCloudDescriptions is None:
            self.aggregateCloud()
        return self._atomCloudDescriptions

    @property
    def residueCloudDescriptions(self):
        """Returns aggregated residue cloud descriptions.

        :return: residueCloudDescriptions
        :rtype: :py:class:`list`
        """
        if self._residueCloudDescriptions is None:
            self.aggregateCloud()
        return self._residueCloudDescriptions

    @property
    def domainCloudDescriptions(self):
        """Returns aggregated domain cloud descriptions.

        :return: domainCloudDescriptions
        :rtype: :py:class:`list`
        """
        if self._domainCloudDescriptions is None:
            self.aggregateCloud()
        return self._domainCloudDescriptions

    @property
    def numVoxelsAggregated(self):
        """Returns number of aggregated voxels in cloud analysis.

        :return: numVoxelsAggregated
        :rtype: :py:class:`int`
        """
        if self._numVoxelsAggregated is None:
            self.aggregateCloud()
        return self._numVoxelsAggregated

    @property
    def totalAggregatedElectrons(self):
        """Returns total amount of aggregated electrons in cloud analysis.

        :return: totalAggregatedElectrons
        :rtype: :py:class:`float`
        """
        if self._totalAggregatedElectrons is None:
            self.aggregateCloud()
        return self._totalAggregatedElectrons

    @property
    def totalAggregatedDensity(self):
        """Returns total amount of aggregated density in cloud analysis.

        :return: totalAggregatedDensity
        :rtype: :py:class:`float`
        """
        if self._totalAggregatedDensity is None:
            self.aggregateCloud()
        return self._totalAggregatedDensity

    @property
    def densityElectronRatio(self):
        """Returns the density-electron ratio estimated from cloud analysis.

        :return: densityElectronRatio
        :rtype: :py:class:`float`
        """
        if self._densityElectronRatio == None:
            self.aggregateCloud()
        return self._densityElectronRatio

    @property
    def atomTypeOverlapCompleteness(self):
        """Returns atom-type overlap completeness counts.

        :return: atomTypeOverlapCompleteness
        :rtype: :py:class:`dict`
        """
        if self._atomTypeOverlapCompleteness is None:
            self.aggregateCloud()
        return self._atomTypeOverlapCompleteness

    @property
    def atomTypeOverlapIncompleteness(self):
        """Returns atom-type overlap incompleteness counts.

        :return: atomTypeOverlapIncompleteness
        :rtype: :py:class:`dict`
        """
        if self._atomTypeOverlapIncompleteness is None:
            self.aggregateCloud()
        return self._atomTypeOverlapIncompleteness



    residueCloudHeader = ['chain', 'residue_number', 'residue_name', 'local_density_electron_ratio', 'num_voxels', 'electrons', 'volume', 'centroid_xyz']
    domainCloudHeader = residueCloudHeader
    def aggregateCloud(self, minCloudElectrons=25.0, minTotalElectrons=400.0):
        """Aggregate the electron density map clouds by atom, residue, and domain.
        Calculate and populate `densityAnalysis.densityElectronRatio` and `densityAnalysis.medians` data members.

        :param minCloudElectrons: minimum number of electrons needed to aggregate a residue or domain cloud., defaults to 25.0
        :type minCloudElectrons: :py:class:`float`
        :param minTotalElectrons: mininum number of electrons required to calculate a density-electron ratio., defaults to 400.0
        :type minTotalElectrons: :py:class:`float`
        """
        densityObj = self.densityObj
        biopdbObj = self.biopdbObj

        domainClouds = []
        domainPool = []
        domainList = []
        residueList = []
        atomList = []

        currentRadii = radiiGlobal
        currentSlopes = slopesGlobal
        completelyOverlappedAtomTypes = collections.defaultdict(int)
        incompletelyOverlappedAtomTypes = collections.defaultdict(int)

        allAtomClouds = {}
        centroidDistances = []
        for residue in biopdbObj.get_residues():
            if residue.id[0] != ' ': # skip HETATOM residues.
                continue

            for atom in residue.child_list:
                resAtom = residueAtomName(atom)
                if resAtom not in fullAtomNameMapAtomTypeGlobal.keys() or atom.get_occupancy() == 0:
                    continue

                atomClouds = densityObj.findAberrantBlobs(atom.coord, currentRadii[fullAtomNameMapAtomTypeGlobal[resAtom]], densityObj.densityCutoff)
                allAtomClouds[tuple(atom.coord)] = atomClouds
                if len(atomClouds) > 0:
                    centroidDistances.append(min(np.linalg.norm(atom.coord - i.centroid) for i in atomClouds))
        centroidDistanceCutoff = np.nanmedian(centroidDistances) + 2.5 * np.nanstd(centroidDistances) # ~99% cutoff, but this is calculated across all atom-types.

        for residue in biopdbObj.get_residues():
            if residue.id[0] != ' ': # skip HETATOM residues.
                continue

            residuePool = []
            atomCloudIndeces = {}
            for atom in residue.child_list:
                resAtom = residueAtomName(atom)
                if resAtom not in fullAtomNameMapAtomTypeGlobal.keys() or atom.get_occupancy() == 0:
                    continue

                ## Calculate atom clouds
                atomClouds = allAtomClouds[tuple(atom.coord)]
                if len(atomClouds) == 0:
                    continue
                elif len(atomClouds) == 1:
                    bestAtomCloud = atomClouds[0]
                else:
                    distances = [np.linalg.norm(atom.coord - i.centroid) for i in atomClouds]
                    minDistance = min(distances)
                    if minDistance > centroidDistanceCutoff:
                        continue
                    index = distances.index(minDistance)
                    bestAtomCloud = atomClouds[index]

                for aCloud in atomClouds:
                    aCloud.atoms = [atom]
                atomCloudIndeces[resAtom] = [len(residuePool)+index for index in range(len(atomClouds))]
                residuePool = residuePool + atomClouds ## For aggregating atom clouds into residue clouds

                atomList.append([residue.parent.id, residue.id[1], atom.parent.resname, atom.name, fullAtomNameMapAtomTypeGlobal[resAtom],
                                 bestAtomCloud.totalDensity / fullAtomNameMapElectronsGlobal[resAtom] / atom.get_occupancy(), len(bestAtomCloud.crsList),
                                 fullAtomNameMapElectronsGlobal[resAtom], atom.get_bfactor(), np.linalg.norm(atom.coord - bestAtomCloud.centroid), bestAtomCloud.centroid])
            ## End atom loop

            overlap = np.zeros((len(residuePool), len(residuePool)))
            for i in range(len(residuePool)):
                for j in range(i+1, len(residuePool)):
                    overlap[i][j] = overlap[j][i] = utils.testOverlap(residuePool[i],residuePool[j])


            ## Calculate atom-type overlap completeness.  Needed for parameter optimization.
            for atom in residue.child_list:
                resAtom = residueAtomName(atom)
                if resAtom in atomCloudIndeces:
                    if all(any(overlap[index1][index2] for index1 in atomCloudIndeces[resAtom] for index2 in atomCloudIndeces[resAtom2]) for resAtom2 in bondedAtomsGlobal[resAtom] if resAtom2 in atomCloudIndeces):
                        completelyOverlappedAtomTypes[fullAtomNameMapAtomTypeGlobal[resAtom]] += 1
                    else:
                        incompletelyOverlappedAtomTypes[fullAtomNameMapAtomTypeGlobal[resAtom]] += 1

            ## Group connected residue density clouds together from individual atom clouds
            resClouds = []
            usedIdx = set()
            for startingIndex in range(len(residuePool)):
                if startingIndex not in usedIdx:
                    newCluster = {index for index, o in enumerate(overlap[startingIndex]) if o}
                    currCluster = set([startingIndex])
                    currCluster.update(newCluster)
                    while len(newCluster):
                        newCluster = {index for oldIndex in newCluster for index, o in enumerate(overlap[oldIndex]) if index not in currCluster and o}
                        currCluster.update(newCluster)

                    usedIdx.update(currCluster)
                    resCloud = residuePool[currCluster.pop()].clone()
                    for idx in currCluster:
                        resCloud.merge(residuePool[idx])
                    resClouds.append(resCloud)

            for cloud in resClouds:
                resElectrons = sum([fullAtomNameMapElectronsGlobal[residueAtomName(atom)] * atom.get_occupancy() for atom in cloud.atoms])
                if resElectrons >= minCloudElectrons:
                    residueList.append([residue.parent.id, residue.id[1], residue.resname, cloud.totalDensity / resElectrons, len(cloud.crsList), resElectrons, len(cloud.crsList) * densityObj.header.unitVolume,
                                        cloud.centroid])

            domainPool = domainPool + resClouds ## For aggregating residue clouds into domain clouds
        ## End residue

        ## Group connected domain density clouds together from individual residue clouds
        overlap = np.zeros((len(domainPool), len(domainPool)))
        for i in range(len(domainPool)):
            for j in range(i+1, len(domainPool)):
                overlap[i][j] = overlap[j][i] = utils.testOverlap(domainPool[i],domainPool[j])

        usedIdx = set()
        for startingIndex in range(len(domainPool)):
            if startingIndex not in usedIdx:
                newCluster = {index for index, o in enumerate(overlap[startingIndex]) if o}
                currCluster = set([startingIndex])
                currCluster.update(newCluster)
                while len(newCluster):
                    newCluster = {index for oldIndex in newCluster for index, o in enumerate(overlap[oldIndex]) if index not in currCluster and o}
                    currCluster.update(newCluster)

                usedIdx.update(currCluster)
                domainCloud = domainPool[currCluster.pop()].clone()
                for idx in currCluster:
                    domainCloud.merge(domainPool[idx])
                domainClouds.append(domainCloud)
        ##End domain

        ## Calculate densityElectronRatio, which is technically a weighted mean value now.
        numVoxels = 0
        totalElectrons = 0
        totalDensity = 0
        for cloud in domainClouds:
            atom = cloud.atoms[0]
            domainElectrons = sum([fullAtomNameMapElectronsGlobal[residueAtomName(atom)] * atom.get_occupancy() for atom in cloud.atoms])
            totalElectrons += domainElectrons
            numVoxels += len(cloud.crsList)
            totalDensity += cloud.totalDensity

            if domainElectrons >= minCloudElectrons:
                domainList.append([atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, cloud.totalDensity / domainElectrons, len(cloud.crsList), domainElectrons, len(cloud.crsList) * densityObj.header.unitVolume,
                                  cloud.centroid])

        if totalElectrons < minTotalElectrons:
            return
        else:
            densityElectronRatio = totalDensity / totalElectrons
            domainList.sort(key=lambda x: x[3])
        ## End calculate densityElectronRatio


        def calcSlope(data, atom_type):
            if len(data['chain']) <= 2 or len(np.unique(data['bfactor'])) == 1: ## Less than three data points or all b factors are the same. Should change this to something more robust like 15 unique values.
                return currentSlopes[atom_type]

            slope, intercept, r_vanue, p_value, std_err = stats.linregress(np.log(data['bfactor']), (data['adj_density_electron_ratio']-densityElectronRatio)/densityElectronRatio)
            return currentSlopes[atom_type] if p_value > 0.05 else slope

        try:
            dataType = np.dtype([('chain', np.dtype(('U', 20))), ('residue_number', int), ('residue_name', np.dtype(('U', 10)) ), ('atom_name', np.dtype(('U', 10)) ), ('atom_type', np.dtype(('U', atomTypeLengthGlobal)) ),
                                 ('density_electron_ratio', float), ('num_voxels', int), ('electrons', int), ('bfactor', float), ('centroid_distance', float), ('centroid_xyz', float, (3,)), ('adj_density_electron_ratio', float),
                                 ('domain_fraction', float), ('corrected_fraction', float), ('corrected_density_electron_ratio', float), ('volume', float)])
            atoms = np.asarray([tuple(atom+[0.0 for x in range(5)]) for atom in atomList],dataType) # must pass in list of tuples to create ndarray correctly.
            if not np.isnan(atoms['centroid_distance']).all():
                centroidCutoff = np.nanmedian(atoms['centroid_distance']) + np.nanstd(atoms['centroid_distance']) * 2
                atoms = atoms[atoms['centroid_distance'] < centroidCutoff]
            atom_types = np.unique(atoms['atom_type'])
            medians = { column : {atom_type : np.nanmedian(atoms[column][atoms['atom_type'] == atom_type]) for atom_type in atom_types} for column in ['num_voxels'] }
            medians_translator = np.vectorize(lambda column,atom_type : medians[column][atom_type])

            ## Normalize by volume (numVoxels)
            atoms['adj_density_electron_ratio'] = atoms['density_electron_ratio'] / atoms['num_voxels'] * medians_translator('num_voxels', atoms['atom_type'])
            atoms['volume'] = atoms['num_voxels'] * densityObj.header.unitVolume
            medians.update({column : {atom_type : np.nanmedian(atoms[column][atoms['atom_type'] == atom_type]) for atom_type in atom_types} for column in
                         ['density_electron_ratio', 'centroid_distance', 'adj_density_electron_ratio', 'volume']})
            medians['bfactor'] = {atom_type : np.nanmedian(atoms['bfactor'][(atoms['atom_type'] == atom_type) & (atoms['bfactor'] > 0)]) for atom_type in atom_types}
            atoms['bfactor'][atoms['bfactor'] <= 0] = medians_translator('bfactor', atoms['atom_type'])[atoms['bfactor'] <= 0]
            medians['slopes'] = {atom_type : calcSlope(atoms[atoms['atom_type'] == atom_type], atom_type) for atom_type in atom_types}

            ## Correct by b-factor
            atoms['domain_fraction'] = (atoms['adj_density_electron_ratio'] - densityElectronRatio) / densityElectronRatio
            atoms['corrected_fraction'] = atoms['domain_fraction'] - (np.log(atoms['bfactor']) - np.log(medians_translator('bfactor', atoms['atom_type']))) * medians_translator('slopes', atoms['atom_type'])
            atoms['corrected_density_electron_ratio'] = atoms['corrected_fraction'] * densityElectronRatio + densityElectronRatio
            medians.update({column : {atom_type : np.nanmedian(atoms[column][atoms['atom_type'] == atom_type]) for atom_type in atom_types} for column in
                         ['domain_fraction', 'corrected_fraction', 'corrected_density_electron_ratio']})
        except:
            return

        self._densityElectronRatio = densityElectronRatio
        self._numVoxelsAggregated = numVoxels
        self._totalAggregatedElectrons = totalElectrons
        self._totalAggregatedDensity = totalDensity
        self._medians = medians
        self._atomCloudDescriptions = atoms
        self._residueCloudDescriptions = residueList
        self._domainCloudDescriptions = domainList
        self._atomTypeOverlapCompleteness = completelyOverlappedAtomTypes
        self._atomTypeOverlapIncompleteness = incompletelyOverlappedAtomTypes


    def medianAbsFoFc(self):
        """Calculates median absolute values for the Fo and Fc maps less than 1 sigma.
        These values should be comparable, i.e. low relative difference, for RSCC and RSR metric calculations.

        :return: tuple of median abs values from fo and fc maps respectively.
        :rtype: :py:class:`tuple`
        """
        fo = self.fo
        fc = self.fc
        foDensityCutoff = fo.meanDensity + 1.0 * fo.stdDensity
        fcDensityCutoff = fc.meanDensity + 1.0 * fc.stdDensity
        ncrs = fo.header.uniqueNcrs
        densityPairs = [ (density, diffDensity) for density, diffDensity in ((utils.getPointDensityFromCrs(fo, crs),utils.getPointDensityFromCrs(fc, crs)) for crs in itertools.product(range(ncrs[0]),range(ncrs[1]),range(ncrs[2])))
          if abs(density) < foDensityCutoff and abs(diffDensity) < fcDensityCutoff ]

        foValues = np.asarray([pair[0] for pair in densityPairs])
        fcValues = np.asarray([pair[1] for pair in densityPairs])
        return (np.median(np.abs(foValues)),np.median(np.abs(fcValues)))

    residueMetricsHeaderList = ['chain', 'residue_number', 'residue_name', "rscc", "rsr", "mean_occupancy", "occupancy_weighted_mean_bfactor"]
    def residueMetrics(self, residueList=None):
        """RETURNS rscc and rsr statistics for each residue using the Fo and Fc density maps.

        :param residueList:
        :type residueList: :py:class:`list`, optional

        :return: results
        :rtype: :py:class:`list`
        """
        resolution = self.biopdbObj.header['resolution']
        radius = 0.7
        if 0.6 <= resolution <= 3:
            radius = (resolution - 0.6) / 3 + 0.7
        elif resolution > 3:
            radius = resolution * 0.5

        if residueList == None:
            residueList = list(self.biopdbObj.get_residues())

        residueResults = []
        for residue in residueList:
            crsList = set()
            bfactorWeightedSum = occupancySum = 0.0
            for atom in residue.child_list:
                crsList.update(utils.getSphereCrsFromXyz(self.fo, atom.coord, radius, 0.0))
                bfactorWeightedSum += atom.get_bfactor() * atom.get_occupancy()
                occupancySum += atom.get_occupancy()

            (rscc, rsr) = self.calculateRsccRsrMetrics(crsList)
            residueResults.append([residue.parent.id, residue.id[1], residue.resname, rscc, rsr, occupancySum / len(residue.child_list), bfactorWeightedSum / occupancySum])

        return residueResults

    atomMetricsHeaderList = ['chain', 'residue_number', 'residue_name', "atom_name", "symmetry", "xyz", "rscc", "rsr", "occupancy", "bfactor"]
    def atomMetrics(self, atomList=None):
        """RETURNS rscc and rsr statistics for each residue using the Fo and Fc density maps.

        :param atomList:
        :type atomList: :py:class:`list`, optional

        :return: results
        :rtype: :py:class:`list`
        """
        resolution = self.biopdbObj.header['resolution']
        radius = 0.7
        if 0.6 <= resolution <= 3:
            radius = (resolution - 0.6) / 3 + 0.7
        elif resolution > 3:
            radius = resolution * 0.5

        if atomList == None:
            atomList = self.asymmetryAtoms

        atomResults = []
        for atom in atomList:
            crsList = set(utils.getSphereCrsFromXyz(self.fo, atom.coord, radius, 0.0))
            (rscc, rsr) = self.calculateRsccRsrMetrics(crsList)
            atomResults.append([atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, atom.name, atom.symmetry, atom.coord, rscc, rsr, atom.get_occupancy(), atom.get_bfactor()])

        return atomResults

    def calculateRsccRsrMetrics(self, crsList):
        """Calculates and returns RSCC and RSR metrics.
        This method of calculating RSCC and RSR assumes that the Fo and Fc maps are appropriately scaled.
        Comparison of median absolute values below one sigma should be quite similar between Fo and Fc maps.

        :param crsList:
        :type crsList: :py:class:`list`, :py:class:`set`

        :return: rscc_rsr_arrays_tuple
        :rtype: :py:obj:`tuple`
        """
        fo = self.fo
        fc = self.fc
        foDensity = np.asarray([utils.getPointDensityFromCrs(fo,i) for i in crsList])
        fcDensity = np.asarray([utils.getPointDensityFromCrs(fc, i) for i in crsList])

        rscc = stats.stats.pearsonr(foDensity, fcDensity)[0]
        rsr = sum(abs(foDensity - fcDensity)) / sum(abs(foDensity + fcDensity))
        return (rscc,rsr)


    def _calculateSymmetryAtoms(self):
        """Calculate all the symmetry and nearby cells and keep those have at least on atom within 5 grid points of the non-repeating crs boundary.
        Ref: Biomolecular Crystallography: Principles, Practice, and Application to Structural Biology by Bernhard Rupp.
        Orthogonalization matrix O and deororthogonalization matrix O' are from :class:`pdb_eda.ccp4` object.
        Rotation matrix R and Translation matrix T is from :class:`pdb_eda.pdbParser` object.
        The neighbouring cells can be calculated using formula,
        X' = O(O'(RX + T) + T') = OO'(RX+T) + OT' = RX + T + O[-1/0/1,-1/0/1,-1/0/1].
        Assign the list of :class:`pdb_eda.densityAnalysis.symAtom` instances to `densityAnalysis.symmetryAtoms` data member
        """
        densityObj = self.densityObj
        biopdbObj = self.biopdbObj
        pdbObj = self.pdbObj

        ## For inRangeAtoms, the min/max range of xyz axes (the circumscribed box)
        ncrs = densityObj.header.ncrs
        orginalDensityBox = [densityObj.header.crs2xyzCoord(i) for i in [[c, r, s] for c in [0, ncrs[0]-1] for r in [0, ncrs[1]-1] for s in [0, ncrs[2]-1]]]
        xs = sorted([i[0] for i in orginalDensityBox])
        ys = sorted([i[1] for i in orginalDensityBox])
        zs = sorted([i[2] for i in orginalDensityBox])

        allAtoms = utils.createSymmetryAtoms(list(biopdbObj.get_atoms()), pdbObj.header.rotationMats, densityObj.header.orthoMat, xs,ys,zs)

        self._symmetryAtoms = allAtoms
        self._symmetryAtomCoords = np.asarray([atom.coord for atom in allAtoms])
        self._symmetryOnlyAtoms = [atom for atom in allAtoms if atom.symmetry != (0,0,0,0)]
        self._symmetryOnlyAtomCoords = np.asarray([atom.coord for atom in self._symmetryOnlyAtoms])
        self._asymmetryAtoms = [atom for atom in allAtoms if atom.symmetry == (0,0,0,0)]
        self._asymmetryAtomCoords = np.asarray([atom.coord for atom in self._asymmetryAtoms])

    blobStatisticsHeader = ['distance_to_atom', 'sign', 'electrons_of_discrepancy', 'num_voxels', 'volume', 'chain', 'residue_number', 'residue_name', 'atom_name', 'atom_symmetry', 'atom_xyz', 'centroid_xyz']
    def calculateAtomSpecificBlobStatistics(self, blobList):
        """Calculate atom-specific blob statistics.

        :param blobList: list of blobs to calculate statistics for.
        :type blobList: :py:class:`list`

        :return blobStats: Difference density map statistics.
        :rtype: :py:class:`list`
        """
        symmetryAtoms = self.symmetryAtoms
        symmetryAtomCoords = self.symmetryAtomCoords

        if not self.densityElectronRatio:
            raise RuntimeError("Failed to calculate densityElectronRatio, probably due to total aggregated electrons less than the minimum.")
        densityElectronRatio = self.densityElectronRatio

        blobStats = []
        for blob in blobList:
            centroid = np.array(blob.centroid).reshape((1, 3))
            symmetryDistances = scipy.spatial.distance.cdist(centroid, symmetryAtomCoords)
            atom = symmetryAtoms[np.argmin(symmetryDistances[0])] # closest atom
            sign = '+' if blob.totalDensity >= 0 else '-'
            blobStats.append([symmetryDistances.min(), sign, abs(blob.totalDensity / densityElectronRatio), len(blob.crsList), blob.volume, atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, atom.name, atom.symmetry, atom.coord, blob.centroid])

        return blobStats

    # Headers that match the order of the results
    regionDensityHeader = [ "actual_significant_regional_density", "num_electrons_actual_significant_regional_density" ]
    atomRegionDensityHeader = ['chain', 'residue_number', 'residue_name', "atom_name", "occupancy"] + regionDensityHeader
    symmetryAtomRegionDensityHeader = ['chain', 'residue_number', 'residue_name', "atom_name", "symmetry", "atom_xyz", "fully_within_density_map"] + regionDensityHeader
    residueRegionDensityHeader = ['chain', 'residue_number', 'residue_name', "mean_occupancy"] + regionDensityHeader


    def calculateAtomRegionDensity(self, radius, numSD=1.5, type=""):
        """Calculates significant region density in a given radius of each atom.

        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance, defaults to 1.5
        :type numSD: py:class:`float`
        :param type: atom type to filter on., defaults to ""
        :type type: :py:class:`str`

        :return diffMapRegionStats: density map region statistics per atom.
        :rtype: :py:class:`list`
        """
        biopdbObj = self.biopdbObj
        atoms = list(biopdbObj.get_atoms())
        if type:
            atoms = [atom for atom in atoms if atom.name == type]

        results = []
        for atom in atoms:
            result = self.calculateRegionDensity([atom.coord], radius, numSD)
            results.append([atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, atom.name, atom.get_occupancy()] + result)

        return results

    def calculateSymmetryAtomRegionDensity(self, radius, numSD=1.5, type=""):
        """Calculates significant region density in a given radius of each symmetry atom.

        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance., defaults to 1.5
        :type numSD: :py:class:`float`
        :param type: atom type to filter on., defaults to ""
        :type type: :py:class:`str`

        :return diffMapRegionStats: density map region statistics per atom.
        :rtype: :py:class:`list`
        """
        atoms = self.symmetryAtoms
        if type:
            atoms = [atom for atom in atoms if atom.name == type]

        results = []
        for atom in atoms:
            (result,valid) = self.calculateRegionDensity([atom.coord], radius, numSD, testValidCrs=True)
            results.append([atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, atom.name, atom.symmetry, atom.coord, valid] + result)

        return results

    def calculateResidueRegionDensity(self, radius, numSD=1.5, type="", atomMask=None):
        """Calculates significant region density in a given radius of each residue.

        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance., defaults to 1.5
        :type numSD: :py:class:`float`
        :param type: atom type to filter on., defaults to ""
        :type type: :py:class:`str`
        :param atomMask: residue specific atom mask.
        :type type: :py:class:`dict`, optional

        :return diffMapRegionStats: density map region statistics per residue.
        :rtype: :py:class:`list`
        """
        biopdbObj = self.biopdbObj

        results = []
        residues = list(biopdbObj.get_residues())
        if type:
            residues = [residue for residue in residues if residue.resname == type]
        for residue in residues:
            atoms = [atom for atom in residue.get_atoms() if not atomMask or (residue.resname in atomMask and atom.name in atomMask[residue.resname])]
            xyzCoordList = [atom.coord for atom in atoms]
            meanOccupancy = np.mean([atom.get_occupancy() for atom in atoms])
            result = self.calculateRegionDensity(xyzCoordList, radius, numSD)
            results.append([residue.parent.id, residue.id[1], residue.resname, meanOccupancy ] + result)

        return results

    def calculateRegionDensity(self, xyzCoordList, radius, numSD=1.5, testValidCrs=False):
        """Calculate region-specific density from the electron density matrix.

        :param xyzCoordLists: single xyz coordinate or a list of xyz coordinates.
        :type xyzCoordList: :py:class:`list`
        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance., defaults to 1.5
        :type numSD: :py:class:`float`
        :param testValidCrs: whether to test crs are valid and return the results., defaults to :py:obj:`False`
        :type testValidCrs: :py:class:`bool`

        :return diffMapRegionStats: density map region statistics and optional validCrs result.
        :rtype: :py:class:`list`, :py:class:`tuple`
        """
        if not self.densityElectronRatio:
            raise RuntimeError("Failed to calculate densityElectronRatio, probably due to total aggregated electrons less than the minimum.")
        densityElectronRatio = self.densityElectronRatio

        densityObj = self.densityObj
        densityCutoff = densityObj.meanDensity + numSD * densityObj.stdDensity

        # observed significant regional density
        blue = densityObj.findAberrantBlobs(xyzCoordList, radius, densityCutoff)
        actual_sig_regional_density = sum([blob.totalDensity for blob in blue])
        num_electrons_actual_sig_regional_density = actual_sig_regional_density / densityElectronRatio

        result = [ actual_sig_regional_density, num_electrons_actual_sig_regional_density ]
        if testValidCrs:
            return (result, utils.testValidXyzList(densityObj, xyzCoordList, radius))
        else:
            return result


    # Headers that match the order of the results
    regionDiscrepancyHeader = [ "actual_abs_significant_regional_discrepancy", "num_electrons_actual_abs_significant_regional_discrepancy",
                 "expected_abs_significant_regional_discrepancy", "num_electrons_expected_abs_significant_regional_discrepancy",
                 "actual_significant_regional_discrepancy", "num_electrons_actual_significant_regional_discrepancy",
                 "actual_positive_significant_regional_discrepancy", "num_electrons_actual_positive_significant_regional_discrepancy",
                 "actual_negative_significant_regional_discrepancy", "num_electrons_actual_negative_significant_regional_discrepancy" ]
    atomRegionDiscrepancyHeader = ['chain', 'residue_number', 'residue_name', "atom_name", "occupancy"] + regionDiscrepancyHeader
    symmetryAtomRegionDiscrepancyHeader = ['chain', 'residue_number', 'residue_name', "atom_name", "symmetry", "atom_xyz", "fully_within_density_map"] + regionDiscrepancyHeader
    residueRegionDiscrepancyHeader = ['chain', 'residue_number', 'residue_name', "mean_occupancy"] + regionDiscrepancyHeader

    def calculateAtomRegionDiscrepancies(self, radius, numSD=3.0, type=""):
        """Calculates significant region discrepancies in a given radius of each atom.

        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance, defaults to 3.0
        :type numSD: py:class:`float`
        :param type: atom type to filter on., defaults to ""
        :type type: :py:class:`str`

        :return diffMapRegionStats: Difference density map region statistics per atom.
        :rtype: :py:class:`list`
        """
        biopdbObj = self.biopdbObj
        atoms = list(biopdbObj.get_atoms())
        if type:
            atoms = [atom for atom in atoms if atom.name == type]

        results = []
        for atom in atoms:
            result = self.calculateRegionDiscrepancy([atom.coord], radius, numSD)
            results.append([atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, atom.name, atom.get_occupancy()] + result)

        return results

    def calculateSymmetryAtomRegionDiscrepancies(self, radius, numSD=3.0, type=""):
        """Calculates significant region discrepancies in a given radius of each symmetry atom.

        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance., defaults to 3.0
        :type numSD: :py:class:`float`
        :param type: atom type to filter on., defaults to ""
        :type type: :py:class:`str`

        :return diffMapRegionStats: Difference density map region statistics per atom.
        :rtype: :py:class:`list`
        """
        atoms = self.symmetryAtoms
        if type:
            atoms = [atom for atom in atoms if atom.name == type]

        results = []
        for atom in atoms:
            (result,valid) = self.calculateRegionDiscrepancy([atom.coord], radius, numSD, testValidCrs=True)
            results.append([atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, atom.name, atom.symmetry, atom.coord, valid] + result)

        return results

    def calculateResidueRegionDiscrepancies(self, radius, numSD=3.0, type="", atomMask=None):
        """Calculates significant region discrepancies in a given radius of each residue.

        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance., defaults to 3.0
        :type numSD: :py:class:`float`
        :param type: atom type to filter on., defaults to ""
        :type type: :py:class:`str`
        :param atomMask: residue specific atom mask.
        :type type: :py:class:`dict`, optional

        :return diffMapRegionStats: Difference density map region statistics per residue.
        :rtype: :py:class:`list`
        """
        biopdbObj = self.biopdbObj

        results = []
        residues = list(biopdbObj.get_residues())
        if type:
            residues = [residue for residue in residues if residue.resname == type]
        for residue in residues:
            atoms = [atom for atom in residue.get_atoms() if not atomMask or (residue.resname in atomMask and atom.name in atomMask[residue.resname])]
            xyzCoordList = [atom.coord for atom in atoms]
            meanOccupancy = np.mean([atom.get_occupancy() for atom in atoms])
            result = self.calculateRegionDiscrepancy(xyzCoordList, radius, numSD)
            results.append([residue.parent.id, residue.id[1], residue.resname, meanOccupancy ] + result)

        return results

    def calculateRegionDiscrepancy(self, xyzCoordList, radius, numSD=3.0, testValidCrs=False):
        """Calculate region-specific discrepancy from the difference density matrix.

        :param xyzCoordLists: single xyz coordinate or a list of xyz coordinates.
        :type xyzCoordList: :py:class:`list`
        :param radius: the search radius.
        :type radius: :py:class:`float`
        :param numSD: number of standard deviations of significance., defaults to 3.0
        :type numSD: :py:class:`float`
        :param testValidCrs: whether to test crs are valid and return the results., defaults to :py:obj:`False`
        :type testValidCrs: :py:class:`bool`

        :return diffMapRegionStats: Difference density map region statistics and optional validCrs result.
        :rtype: :py:class:`list`, :py:class:`tuple`
        """
        if not self.densityElectronRatio:
            raise RuntimeError("Failed to calculate densityElectronRatio, probably due to total aggregated electrons less than the minimum.")
        densityElectronRatio = self.densityElectronRatio

        diffDensityObj = self.diffDensityObj
        diffDensityCutoff = diffDensityObj.meanDensity + numSD * diffDensityObj.stdDensity

        # observed significant regional discrepancy
        green = diffDensityObj.findAberrantBlobs(xyzCoordList, radius, diffDensityCutoff)
        red = diffDensityObj.findAberrantBlobs(xyzCoordList, radius, -1.0 * diffDensityCutoff)
        actual_positive_sig_regional_discrep = sum([blob.totalDensity for blob in green])
        num_electrons_actual_positive_sig_regional_discrep = actual_positive_sig_regional_discrep / densityElectronRatio
        actual_negative_sig_regional_discrep = sum([blob.totalDensity for blob in red])
        num_electrons_actual_negative_sig_regional_discrep = actual_negative_sig_regional_discrep / densityElectronRatio
        actual_sig_regional_discrep = actual_positive_sig_regional_discrep + actual_negative_sig_regional_discrep
        num_electrons_actual_sig_regional_discrep = actual_sig_regional_discrep / densityElectronRatio
        actual_abs_sig_regional_discrep = abs(actual_positive_sig_regional_discrep) + abs(actual_negative_sig_regional_discrep)
        num_electrons_actual_abs_sig_regional_discrep = actual_abs_sig_regional_discrep / densityElectronRatio

        # expected absolute significant regional discrepancy
        total_abs_sig_discrep = diffDensityObj.getTotalAbsDensity(diffDensityCutoff)
        total_voxel_count = len(diffDensityObj.densityArray)
        avg_abs_vox_discrep = total_abs_sig_discrep / total_voxel_count
        regional_voxel_count = len(utils.getSphereCrsFromXyzList(diffDensityObj, xyzCoordList, radius))
        expected_abs_sig_regional_discrep = avg_abs_vox_discrep * regional_voxel_count
        num_electrons_expected_abs_sig_regional_discrep = expected_abs_sig_regional_discrep / densityElectronRatio

        result = [ actual_abs_sig_regional_discrep, num_electrons_actual_abs_sig_regional_discrep,
                 expected_abs_sig_regional_discrep, num_electrons_expected_abs_sig_regional_discrep,
                 actual_sig_regional_discrep, num_electrons_actual_sig_regional_discrep,
                 actual_positive_sig_regional_discrep, num_electrons_actual_positive_sig_regional_discrep,
                 actual_negative_sig_regional_discrep, num_electrons_actual_negative_sig_regional_discrep ]

        if testValidCrs:
            return (result, utils.testValidXyzList(diffDensityObj, xyzCoordList, radius))
        else:
            return result


    def estimateF000(self):
        """Estimate the F000 term as sum of all electrons over the unit cell volume

        :return: estimatedF000
        :rtype: :py:class:`float`
        """
        densityObj = self.densityObj
        biopdbObj = self.biopdbObj
        pdbObj = self.pdbObj

        if not elementElectronsGlobal:
            loadF000Parameters()

        totalElectrons = 0

        allAtoms = list(biopdbObj.get_atoms())
        for atom in allAtoms:
            fullAtomName = residueAtomName(atom)
            if fullAtomName in masterFullAtomNameMapElectronsGlobal:
                totalElectrons += masterFullAtomNameMapElectronsGlobal[fullAtomName]
            elif atom.element in elementElectronsGlobal:
                totalElectrons += elementElectronsGlobal[atom.element] + 1 # + 1 as an estimate for the number of H.

        totalElectrons *= len(pdbObj.header.rotationMats)
        asuVolume = densityObj.header.unitVolume * densityObj.header.nintervalX * densityObj.header.nintervalY * densityObj.header.nintervalZ

        return totalElectrons/asuVolume


def residueAtomName(atom):
    """Returns a combined residue and atom name used to select an atom type.

    :param atom:
    :type atom: :class:`Bio.PDB.atom`

    :return: fullAtomName
    :rtype: :py:class:`str`
    """
    return atom.parent.resname.strip() + '_' + atom.name